// s2255demo.c
// Linux Demo Application for 2255.  Does not require V4L
//
// Copyright (C) 2007-2011  Sensoray Co., Inc.
//
// @description: The purpose of this demo application is to demonstrate the capabilities of the 2255.
// This demo allows viewing the stream(s) with XWindows and saving snapshots.
// The end user can manipulate the buffers as desired(display to SDL/OpenGL/XWindows,etc.)
// A note on Video for Linux (V4L or V4L2): (see s2255.c)
//
// COPY_METHOD involves double buffering(inefficient)
// It is highly recommend to use mmap instead of COPY_METHOD
//#define COPY_METHOD 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <pthread.h>
#include <sched.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
// X stuff
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
// shared memory (for image stuff)
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include <assert.h>
#include <unistd.h>
#include "s2255.h"
#include "s2255f.h"
#include "s2255ioctl.h"
#include "image.h"

#ifndef S2255_64BIT
#define _grayscale_rgb565 grayscale_rgb565_mmx
#else
#define _grayscale_rgb565 grayscale_rgb565
#endif
// Manifest constants ////////////////////////////////////////////////////////
#define TRUE                1
#define FALSE               0

#define MAXLEN              80         // Maximum length of user command string.
#define NUMFRAMES           SYS_FRAMES

// Stream mode:
#define MODE_IDLE           0          // Halted.
#define MODE_STREAM         1          // Streaming
#define MODE_REC            2          // recording
#define MODE_STREAM_TEST    3          // Streaming test for QA only(no display)
// Configuration constants ///////////////////////////////////////////////////
// max number of 2255 boards in the system to look for(1 to 8).
#define NUM_BOARDS          8

static int G_numboards = 0; // number of boards detected

// Storage ///////////////////////////////////////////////////////////////

typedef struct BOARD {                          // BOARD OBJECT -------
    int             handle;                     // Handle to 2255 device.(file open handle)
    int             smode[MAX_CHANNELS];        // Stream mode.
    int             stop_rq[MAX_CHANNELS];      // Stream Halt requested.
    char            fname[MAXLEN];              // File name for snapshots
    pthread_t       io_thread_id[MAX_CHANNELS]; // Handle to stream i/o thread.
    unsigned int    frame_count[MAX_CHANNELS];  // Number of frames transferred
    unsigned int    drop_count[MAX_CHANNELS];   // Number of frames dropped
    int             b_snapshot[MAX_CHANNELS];   // trigger snapshot
    int             flags[MAX_CHANNELS];        // Status flags from i/o to main thread.
    MODE2255        mode[MAX_CHANNELS];         // 2255 Mode structure
    BUFFER          m_buf[MAX_CHANNELS];        // buffer data
    int             board;                      // board number
} BOARD;

static struct BOARD bd[NUM_BOARDS];     // Board objects.
static int          act_bdnum   = 0;    // Board 0 is active by default.


// Function prototypes ////////////////////////////////////////////////////

static int app_init( void );
static void app_shutdown( void );
static void cmd_processor( void );
static void *io_thread_func( void* arg );
static void get_attributes( struct BOARD *b );
static void start_stream( struct BOARD *b, int chan, int new_stream_mode );
static void stop_stream( struct BOARD *b, int chan );
// X Windows
static Window createWindow(Display** display, GC *gc, int width, int height,int b, int c);
static void closeWindow(Display* display, Window win, GC gc);


///////////////////////////////////////////////////////////////////////////
// Init the system and execute user commands until termination.

int main( int argc, char *argv[] )
{
    int rtnval;

    // Attempt to open and initialize all boards.
    rtnval = app_init();

    // If all boards successfully opened and initialized ...
    if (rtnval == 0 ) {
        // Execute user commands until Application Quit command is invoked.
        cmd_processor();
    }

    // Close all opened boards.
    app_shutdown();

    // Exit application.
    return 0;
}

///////////////////////////////////////////////////////////////////////
// Open and initialize one board.  Returns: 0=success, -1=error.

static int board_open( int board_num )
{
    // Set pointer to board's object.
    struct BOARD *b = &bd[board_num];
    MODE2255 defmode ={DEF_MODE_NTSC_CONT};
    int ret;
    int chn;
    int i;
    ret = S2255_DeviceOpen(board_num, &b->handle);
    if (ret != 0) {
        printf("could not open board.\n");
        printf("Note: on some Linux versions, device could take up to 6 seconds to register the USB device\n");
        printf("If driver was just loaded, please try again after waiting\n");
        return ret;
    }
    if (b->handle == -1) {
        return -1;
    }

    defmode.color = COLOR_YUVPL;
    for( chn = 1 ;chn<=MAX_CHANNELS;chn++) {
        ret = S2255_SetMode(b->handle, chn, &defmode);
        if (ret != 0 ) {
            printf("failed set mode 0x%x\n",ret);
            return ret;
        }
        b->mode[chn-1] = defmode;
    }
    // Initialize object.
    for(i = 0;i<MAX_CHANNELS;i++) {
        b->smode[i] = MODE_IDLE;
        b->flags[i]   = 0;
    }
    b->fname[0] = 0;
    b->board = board_num;

    if (b->handle == -1 ) {
        printf( "Driver open error\n" );
        return( -1 );
    }
    return 0;
}

///////////////////////////////////////////////////////////////////////
// Open and initialize all boards.  Returns: -1=error, 0=success.

static int app_init( void )
{
    int     i;
    // For each board ...
    for ( i = 0; i < NUM_BOARDS; i++ )
    {
        // Open device.  Return error if open failed.
        if (board_open( i ) == -1 ) {
            return G_numboards ? 0 : -1;
        }
        G_numboards++;
    }

    return 0;
}

//////////////////////////////////////////////////////////////
// Terminate all stream threads and close all boards.

static void app_shutdown( void )
{
    struct BOARD *b;
    int i;
    // For each board ...
    for ( b = bd; b < &bd[NUM_BOARDS]; b++ )
    {
        // We are finished if this board was never opened.
        if (b->handle == -1 )
            break;

        // Terminate the board's I/O thread if it exists.
        for( i=1;i<=MAX_CHANNELS;i++) {
            stop_stream( b,i );
        }

        // Close the board.
        close( b->handle );
    }

    printf( "\n" );
}

//////////////////////////////////////////////////////////////
// Display the user prompt.

static void prompt( void )
{
    printf( "s2255demo[%d]>", act_bdnum );
    fflush( stdout );
}

//////////////////////////////////////////////////////////////
// Display the command menu.

static void show_menu( void )
{
    printf( "+-S2255DEMO-COMMANDS------------------------------------------------------------+\n" );
    printf( "|   board n ....... select board: 0..7                                         |\n" );
    printf( "|   get ........... display current settings                                   |\n" );
    printf( "|   h[elp] ........ display this list                                          |\n" );
    printf( "|   q[uit] ........ quit program                                               |\n" );
    printf( "|   getSN ......... read the serial number from the eeprom                     |\n" );
    printf( "|   getFW ......... reads the USB and DSP firmware versions                    |\n" );
    printf( "+-VIDEO------------------------------------------------------------------------+\n" );
    printf( "|   bright c v .... set luminance brightness v: -127..127(def 0) for channel c |\n" );
    printf( "|   contrast c v ...set luminance contrast v: 0..255 for channel c             |\n" );
    printf( "|   vsat c v ...... set chrominance saturation v: 0..255 for chanenl c         |\n" );
    printf( "|   vsystem c v.... set video system v: 1=NTSC, 2=PAL for channel c            |\n" );
    printf( "|   fdec c v....... set frame decimation c=channel, v=value(1 all frames,      |\n" );
    printf( "|                   2 every 2nd, 3 every 3rd, 5 every fifth)                   |\n" );
    printf( "|   scale c v...... set scale c=channel, v=value(1-4CIFS, 2-2CIFS, 3-1CIFS,    |\n" );
    printf( "|                   4-4CIFSI(full size, de-interlaced/interpolated), 5-1CIFSF   |\n" );
    printf( "|   color c v...... set color c=channel, v=value(1-color(default), 0 grayscale)|\n" );
    printf( "|   intype c t..... set video type for 2257(svideo=1, composite=0)             |\n" );
    printf( "|   status c ...... get video status for channel c=channel                     |\n" );
    printf( "+-STREAM-----------------------------------------------------------------------+\n" );
    printf( "|   start c........ start preview stream c = channel(1,2,3,4)                  |\n" );
    printf( "|   rec c f........ start record raw stream c = channel(1,2,3,4) f=file+ext    |\n" );
    printf( "|   IMPORTANT NOTE: \"rec\" is only to save the raw stream only(see README).     |\n");
    printf( "|   stop c......... stop preview stream c = channel(1,2,3,4)                   |\n" );
    printf( "|   snap c f....... takes snapshot on channel c = channel(1,2,3,4) to file f   |\n" );    
    printf( "+------------------------------------------------------------------------------+\n" );
}

///////////////////////////////////////////////////////////////////////////////
// Decode and execute user command.  Returns TRUE if Quit selected.

static int exec_command( char *cmd, char *arg,char *arg2 )
{
    struct BOARD    *b;

    // Convert argument to integer value (not relevant for all commands).
    int arg_val = -1;
	int valid_channel; // used only for channel commands
    sscanf( arg, "%d", &arg_val );
	valid_channel = (arg_val >= 1) && (arg_val <= 4);

    // Decode and execute command -----------------------------------

    // Terminate application.
    if (cmd[0] == 'q' )
        return TRUE;

    // Select board.
    if (strcmp( cmd, "board" ) == 0 ) {
        if (( arg_val >= 0 ) && ( arg_val < G_numboards ) ) {
            act_bdnum = arg_val;
        }
        else {
            printf( "Illegal or missing argument; must be in range 0..%d\n", G_numboards-1);
            prompt();
            return FALSE;
        }
        prompt();
        return FALSE;
    }
    // Show help.
    if (( cmd[0] == 'h' ) || ( cmd[0] == '?' ) ) {
        show_menu();
        prompt();
        return FALSE;
    }
    // Prompt for user command.
    prompt();
    // Set pointer to active board's object.
    b = &bd[act_bdnum];
    if (strcmp( cmd, "get"    ) == 0 ) {
        get_attributes(b);
    }
    else if (strcmp( cmd, "start"       ) == 0 ) {
		if (arg[0] == 0) {
			start_stream( b, 1, MODE_STREAM);
		} else if (valid_channel) {
			start_stream( b, arg_val, MODE_STREAM);
		} else {
			printf("invalid channel number\n");
			prompt();
		}
    }
    else if (strcmp( cmd, "rec"       ) == 0 ) {
        strcpy(b->fname, ( arg2 == 0 ) ? "raw.yuv" : arg2);
		if (arg[0] == 0) {
			start_stream(b, 1, MODE_REC);
		} else if (valid_channel) {
			start_stream(b, arg_val, MODE_REC);
		} else {
			printf("invalid channel number\n");
			prompt();
		}
    }    
    else if (strcmp( cmd, "stop"      ) == 0 ) {
		if (arg[0] == 0) {
			stop_stream(b, 1);
		} else if (valid_channel) {
			stop_stream(b, arg_val);
		} else {
			printf("invalid channel number\n");
			prompt();
		}
    }
    else if (strcmp( cmd, "snap"       ) == 0 ) {
        int chn,val;
        sscanf( arg, "%d",&chn);
		if (!valid_channel) {
			printf("invalid channel number\n");
			prompt();
		} else {
			strcpy(b->fname, ( arg2[0] == 0 ) ? "snapshot.bmp" : arg2);
			if (b->smode[chn-1] == MODE_IDLE) {
				printf("stream must be running !\n");
				prompt();
			} else {
				// fire snapshot in IO thread
				b->b_snapshot[chn-1] = 1;
			}
		}
    }
    else if (strcmp( cmd, "fdec" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn >4 || chn < 1 || val <1 || val > 5 || val == 4) {
            printf("bad channel or decimation value %d %d\n",chn,val);
            prompt();
        } else {
            int ret;
            b->mode[chn-1].fdec = val;
            ret = S2255_SetMode( b->handle, chn, &b->mode[chn-1]);
            if (ret != 0 ) {
                printf("failed set mode 0x%x\n", ret);
                prompt();
            }
        }
    }
    else if (strcmp( cmd, "scale" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn >4 || chn < 1 || val <1 || val > 5) {
            printf("bad channel or scale value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            /* You can't change the scale on the fly */
            if (b->smode[chn-1] != MODE_IDLE) {
                printf("you must stop the stream first\n");
                prompt();
            } else {
                b->mode[chn-1].scale = val;
                ret = S2255_SetMode( b->handle, chn, &b->mode[chn-1]);
                if (ret != 0 ) {
                    printf("failed set mode\n");
                    prompt();
                }
            }
        }
    }  
    else if (strcmp( cmd, "bright" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
         if (chn >4 || chn < 1 || val < -127 || val >= 128) {
            printf("bad channel or bright value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            b->mode[chn-1].bright = val;
            ret = S2255_SetMode( b->handle, chn, &b->mode[chn-1]);
            if (ret != 0 ) {
                printf("failed set mode\n");
                prompt();
            }
        }
    }
    else if (strcmp( cmd, "contrast" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn >4 || chn < 1 || val <0 || val > 255) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            b->mode[chn-1].contrast = val;
            ret = S2255_SetMode( b->handle, chn, &b->mode[chn-1]);
            if (ret != 0 ) {
                printf("failed set mode\n");
                prompt();
            }
        }
    }
    else if (strcmp( cmd, "vsat" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn >4 || chn < 1 || val <0 || val > 255) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            b->mode[chn-1].saturation = val;
            ret = S2255_SetMode( b->handle, chn, &b->mode[chn-1]);
            if (ret != 0 ) {
                printf("failed set mode\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "color" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn >4 || chn < 1 || val <0 || val > 1) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            b->mode[chn-1].color &= ~MASK_COLOR;
            b->mode[chn-1].color |= val ? COLOR_YUVPL : COLOR_Y8;
            ret = S2255_SetMode( b->handle, chn, &b->mode[chn-1]);
            if (ret != 0 ) {
                printf("failed set color\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "intype" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn >4 || chn < 1 || val <0 || val > 1) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            b->mode[chn-1].color &= ~MASK_INPUT_TYPE;
            b->mode[chn-1].color |= val ? (1 << 16) : 0;
            ret = S2255_SetMode( b->handle, chn, &b->mode[chn-1]);
            if (ret != 0 ) {
                printf("failed set color\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "vsystem" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn >4 || chn < 1 || val <1 || val > 2) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        }
        else {
            if (b->smode[chn-1] != MODE_IDLE) {
                printf("you must stop the stream first\n");
                prompt();
            } else {
                int ret;
                b->mode[chn-1].format = val; 
                ret = S2255_SetMode( b->handle, chn, &b->mode[chn-1]);
                if (ret != 0 ) {
                    printf("failed set video system\n");
                    prompt();
                }
            }
        }
    } else if (strcmp( cmd, "status" ) == 0) {
        int chn;
        int status, ret;
        sscanf( arg, "%d",&chn);
        // check the channel
       if (!valid_channel) {
            printf("bad channel or value %d\n",chn);
            prompt();
        } else {
            ret = S2255_get_vid_status( b->handle, chn, &status);
            if (ret != 0 ) {
                printf("failed get video status: %d\n", ret);
                prompt();
            } else {
                printf("got status %x\n", status);
                if ((status & 0x01) == 0x00) {
                    printf("no video signal detected\n");
                    prompt();
                } else {
                    printf("video signal present: ");
                    if ((status & 0x0100) == 0x0100) {
                        printf("PAL(or other 50Hz signal)\n");                  
					}
					else {
                        printf("NTSC(or other 60Hz signal)\n");
                    }
                    prompt();
                }
            }
        }
    } else if (strcmp( cmd, "getsn" ) == 0) {
	    SN2255 sn;
	    int res;
	    
	    res = S2255_GetSN(b->handle, &sn);
	    
	    if (res) {
		    printf("Could not query SN, error code: %d\n", res);
		    prompt();
	    }
	    else {
		    printf("\nSerial Number: %d\nDate Programmed: %02d-%02d-%02d\n", sn.serial_number, sn.month, sn.day, sn.year);		
		    prompt();
	    }
    } else if (strcmp( cmd, "getfw" ) == 0) {
	    unsigned int usb, dsp;
	    int res;
	    res = S2255_GetFirmware(b->handle, &usb, &dsp);
	    
	    if (res) {
		    printf("Could not query firmware, error code: %d\n", res);
		    prompt();
	    }
	    else {
		    printf("\nUSB version: %d.%d, DSP FW %d.%d.%d\n",
			   (usb >> 8) & 0xff,
			   (usb & 0xff), dsp / 10000, (dsp % 10000) /100, (dsp % 100));
		    prompt();
	    }
    } else if (cmd[0] != 0 ) {
	    printf( "Invalid command: %s\n", cmd );
	    prompt();
    }
    return FALSE;
}

////////////////////////////////////////////////////////////////////
// Execute user commands until Application Quit is requested.

static void cmd_processor( void )
{
    char                    cmdbuf[MAXLEN];
    char                    cmd[MAXLEN];
    char                    arg[MAXLEN];
    char                    arg2[MAXLEN];
    char                    *c;
    int                 arg_val;

    // Display the command menu.
    show_menu();

    // Prompt for user command.
    prompt();

    // Repeat until Application Terminate is requested ...
    for ( ; ; )
    {
        // Input user command string, and if no errors encountered ...
        if (fgets( cmdbuf, MAXLEN - 1, stdin ) != NULL )
        {
            // Truncate newline char at end of command string.
            cmdbuf[ strlen( cmdbuf ) - 1 ] = 0;

            // If command line is not blank ...
            if (cmdbuf[0] != 0 )
            {
                // Force command string to lower case, except command=="text".
                if (strncmp( cmdbuf, "text", 4 ) != 0 )
                    for ( c = cmdbuf; *c != 0; c++ ) 
                        *c = tolower( *c );

                // Parse command string into opcode and argument.
                cmd[0] = arg[0] = 0;
                sscanf( cmdbuf, "%s %s %s", cmd, arg,arg2 );
                // Decode and execute command
                if (exec_command( cmd, arg,arg2 ) )
                    break;
            }
            else
            {
                // Prompt for user command.
                prompt();
            }
        }
    }
}

/////////////////////////////////////////////////////
// Fetch and display board's current settings.
static char txt_vformat[][40] = {"", "NTSC", "PAL"};


// Fetch and display the attributes.
static void get_attributes( struct BOARD *b )
{
    int i;
    printf( "\nCURRENT SETTINGS -----------\n" );
    for( i=0;i<MAX_CHANNELS;i++) {
        printf( "\nBoard %d----------\n", i );  
        printf( "video format(system)  = %d %s\n", b->mode[i].format, txt_vformat[b->mode[i].format]);
    }
}

struct io_thread_param {
    BOARD   *b;
    int     chn;
};

static void start_stream( struct BOARD *b, int chan, int new_stream_mode )
{
    pthread_attr_t attr;
    struct io_thread_param *param;
    // If we are already streaming ...
    if (b->smode[chan-1] != MODE_IDLE ) {
        // Abort command.
        printf( "already streaming\n" );
        prompt();
        return;
    }

    // Indicate the new stream mode.
    b->smode[chan-1] = new_stream_mode;

    // Clear the stream thread's termination request.
    b->stop_rq[chan-1] = 0;
    // Create detached thread that will manage the stream flow.
    pthread_attr_init( &attr );
    pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED );
    param = (struct io_thread_param *) malloc( sizeof(struct io_thread_param));
    if (param == NULL) {
        printf("out of memory \n");
        return;
    }
    param->b = b;
    param->chn = chan;

    pthread_create(&b->io_thread_id[chan-1], NULL, io_thread_func, param );
    pthread_attr_destroy( &attr );
}

//////////////////////////////////////////////////
// Halt data stream

static void stop_stream( struct BOARD *b, int chan )
{
    // If stream i/o thread is running ...
    if (b->smode[chan-1] != MODE_IDLE )
    {
        // Request thread termination.
        b->stop_rq[chan-1] = 1;
        // Wait for the thread to terminate
        pthread_join( b->io_thread_id[chan-1], NULL);

    }
}


////////////////////////////////////////////////////////////////////
// I/O THREAD MAIN.  Read capture frames until stream is stopped.
void *io_thread_func( void *arg )
{
    time_t                 starttime;                       // Stream start time.
    BOARD                  *b;
    int                    i;
    int                    idx = 0;
    int                    chn = 2;
    int                    active_frame = 0; // active frame count
    unsigned char          *rgb_frame; // for snapshots
    LPBITMAPINFO           lpbmi;    
    struct io_thread_param *param = (struct io_thread_param *) arg;
    UINT32                 size;
    int                    drv_frame;
    int                    previous_frame;
    image_info_t           info;
    int                    depth;
    int                    width;
    int                    height;
    int                    bpp;
    FILE                   *curfile;
    Display *display = NULL;
    char *work_image = NULL;
    GC gc;
    Window windowId;
    int screenNum;
    XImage* ximage = NULL;
    XShmSegmentInfo shminfo;
    Pixmap pix;
    Atom wmDeleteMessage;
    XEvent event;
    int k;
    int nf = 0; // no frame count
    b = (BOARD *) param->b;
    chn = param->chn;
    idx = chn - 1;
    
    // Initialize stream statistics.
    b->frame_count[idx] = 0;
    b->drop_count[idx] = 0;
    time( &starttime );     // Log the start time.

    if (b->smode[idx] == MODE_REC) {
        // open the file
        curfile = fopen( b->fname, "wb+");
        if (curfile == NULL) {
            printf("error opening file!\n");
            prompt();
            return NULL;
        }
        /* allocate large enough image for YUY2 pal or NTSC */
        work_image = (char *) malloc(576*768*2);
        if (work_image == NULL) {
            printf("out of memory\n");
            return NULL;
        }
    }
        
    if (b->handle == 0) {
        printf("device not open. please restart!\n");
    }

    // allocate space for NUMFRAMES frames
    memset(&b->m_buf[idx], 0, sizeof(b->m_buf[idx]));
    b->m_buf[idx].dwFrames = NUMFRAMES;

    /* allocate buffers large enough to allow color switching while streaming
     * largest image size will be PAL  RGB
     * your own application may choose to use smaller buffers if the
     *mode (NTSC/PAL, color/monochrome, size(1CIF, 2CIF) is constant
     */
    size = 576*768*3;
    b->m_buf[idx].size = size;

#ifndef COPY_METHOD
    // MMAP
    for( i=0;i<NUMFRAMES;i++) {
        getbuf_param_t cmd;        
        cmd.idx = idx;
        cmd.frm_num = i;
        ioctl( b->handle,S2255_IOC_REQBUF, &cmd);
        //printf("cmd pdata %p\n", cmd.pbuf);
        b->m_buf[idx].frame[i].pdata = (char *)mmap( 0, size, PROT_READ, MAP_PRIVATE, b->handle, (off_t) cmd.pbuf);//vidbuf->offset);
        if (b->m_buf[idx].frame[i].pdata == NULL) {
            printf("mmap failed\n");
            free(arg);
            return NULL;
        }
        //printf("mmap mem %p\n", b->m_buf[idx].frame[i].pdata);
    }
#else
    // inefficient.  copy buffers 
    for( i=0;i<NUMFRAMES;i++) {
        b->m_buf[idx].frame[i].pdata = (char *) malloc(size);
        memset( b->m_buf[idx].frame[i].pdata,0xcd,size);
        //printf("size %d\n", size);
        if (size == 0) {
            free( arg);
            return NULL;
        }
    }
#endif
    
    active_frame = 0;
    if (S2255_StartAcquire( b->handle, chn, NULL) != 0) {
        printf("Start Acquire Failed");
        b->m_buf[idx].size = 0;        
        for( i=0;i<NUMFRAMES;i++) {
            if (b->m_buf[idx].frame[i].pdata ) {
                free( b->m_buf[idx].frame[i].pdata );
                b->m_buf[idx].frame[i].pdata = NULL;
            }            
        }
        free( arg);     
        return NULL;
    }
    prompt();


    rgb_frame = (unsigned char *)  malloc( 576*720*3);
    CreateDIBHeader( &b->mode[idx],&lpbmi);
    width = lpbmi->bmiHeader.biWidth;
    height = lpbmi->bmiHeader.biHeight;

    // Report status.
    printf( "\nStarting stream on channel %d\n",chn);
    prompt();

    if (b->smode[idx] == MODE_STREAM) {

        //================================================
        // BEGIN code to create window
        // Create the window
        windowId = createWindow(&display, &gc, width, height, b->board, chn);
        // Retrieve the screen color depth
        screenNum = DefaultScreen(display);
        //Screen* screen = ScreenOfDisplay(display, screenNum);
        depth = DefaultDepth(display, screenNum);
        // Create the XImage needed to display the video
        memset(&shminfo, 0, sizeof(XShmSegmentInfo));
        ximage = XShmCreateImage(display, DefaultVisual(display, screenNum), depth,
                                 ZPixmap, NULL, &shminfo, width,height); 
        if (ximage == NULL) {
            printf("Couldn't create XImage.\n");
            exit(1);
        }
        // X windows depth and X image BPP is different
        if (depth < 16) {
            if (ximage->bits_per_pixel == 16) {
                bpp = 16;
            }
            else {
                printf("invalid display format!\n");
                bpp = ximage->bits_per_pixel;
            }
        }
        else {
            // Xwindows.  actual bitmap can be either RGB24 or RGB32
            // even though screen depth 24
            bpp = ximage->bits_per_pixel;
        }
        printf("Xwindows depth %d, BPP %d\n", depth, bpp);
        prompt();
        shminfo.shmid = shmget(IPC_PRIVATE,
                               ximage->bytes_per_line * ximage->height,
                               IPC_CREAT | 0777);
        if (shminfo.shmid == -1) {
            printf("Couldn't allocate shared memory.\n");
            exit(1);
        }
        shminfo.shmaddr = (char*)shmat(shminfo.shmid, 0, 0);
        if (shminfo.shmaddr == (void*)-1) {
            printf("Couldn't attach shared memory.\n");
            exit(1);
        }
        ximage->data = shminfo.shmaddr;
        shminfo.readOnly = False;
        XShmAttach(display, &shminfo);
        XSync(display, False);
        shmctl(shminfo.shmid, IPC_RMID, NULL);
        // Create the pixmap
        pix = XCreatePixmap(display, windowId, width, height, depth);
        // register close event
        wmDeleteMessage = XInternAtom(display, "WM_DELETE_WINDOW", False);
        XSetWMProtocols(display, windowId, &wmDeleteMessage, 1);
        // end code to create window
    }
    previous_frame = -1;
    // Until stream termination is requested ...
    while ( !b->stop_rq[idx] )
    {
        int state[NUMFRAMES]; // state of individual frames (0-idle, 1-filling(in progress), 2-ready)
        int ts[NUMFRAMES];    // timestamp of individual frames(if frame ready)
        // check if frame ready.  Call is non-blocking
        if (S2255_check_frame_poll(b->handle, chn, &drv_frame, NULL) != 0) {
            struct timespec tv;
            tv.tv_sec = (time_t)0;
            tv.tv_nsec = 1000*1000;
            nanosleep( &tv, &tv );
            nf++;
            if (nf > 100) {
                int frame_ready = 0;
                nf = 0;
                printf("no frame in 100ms\n");
                (void) S2255_query_buf_all(b->handle, chn, state, ts);
                for (k = 0; k < NUMFRAMES; k++) {
                    if (state[k] != 0) {
                        // dq it
                        S2255_DQBUF( b->handle, chn, k);
                    }
                }
                // debug printout
//              printf("%d %d %d %d %d %d: ", state[0], state[1], state[2], state[3], state[4], state[5]);
                
//              printf("%d %d %d %d %d %d\n", ts[0], ts[1], ts[2], ts[3], ts[4], ts[5]);
            }
            continue;
        }
        nf = 0;
        previous_frame = drv_frame;

        if (drv_frame != active_frame) {
            int cur_frame;
            while(active_frame != drv_frame) {
                printf("dropping frame\n");
                S2255_DQBUF( b->handle, chn, active_frame);
                // also count down the semaphore in the driver
                S2255_check_frame_poll(b->handle, chn, &cur_frame, NULL);
                // increment number of dropped frames
                b->drop_count[idx]++;
                active_frame++;                
                if (active_frame >= NUMFRAMES) {
                    active_frame = 0;
                }
            }
        }
        
#ifdef COPY_METHOD
        // get this frame.  not necessary if using mmap, which you should.
        if (S2255_get_buffer(b->handle, chn,
                             drv_frame,
                             b->m_buf[idx].frame[drv_frame].pdata,
                             b->m_buf[idx].size) != 0) {
            //printf("get buffer failed\n");
            continue;
        }
        //printf("last frame success %d\n", drv_frame);        
#endif
        // do work with frame here.
        // For our demo, we either display the image to X
        // OR take a snapshot
        if (b->smode[idx] == MODE_REC) {
            /* convert to YUY2 for mplayer playback */
            planar422p_to_yuy2(  b->m_buf[idx].frame[drv_frame].pdata, 
                                 work_image,
                                 lpbmi->bmiHeader.biWidth,
                                 lpbmi->bmiHeader.biHeight);

            fwrite( work_image, 
                    1,
                    lpbmi->bmiHeader.biWidth * lpbmi->bmiHeader.biHeight * 2,
                    curfile);
        }
        else if (b->smode[idx] == MODE_STREAM) {
            if (b->b_snapshot[idx] == 0) {
                if (bpp == 16) {
                    // Xwindows depth 16 RGB565
                    if ((b->mode[idx].color & MASK_COLOR) == COLOR_YUVPL) {
                        planar422p_to_rgb565((const unsigned char *) b->m_buf[idx].frame[drv_frame].pdata, 
                                             (unsigned char *)ximage->data, //rgb_frame,
                                             lpbmi->bmiHeader.biWidth,
                                             lpbmi->bmiHeader.biHeight);
                    }
                    else if ((b->mode[idx].color & MASK_COLOR) == COLOR_Y8) {
                        _grayscale_rgb565((const unsigned char *) b->m_buf[idx].frame[drv_frame].pdata, 
                                             (unsigned char *)ximage->data, //rgb_frame,
                                             lpbmi->bmiHeader.biWidth,
                                             lpbmi->bmiHeader.biHeight);                    
                    }
                }
                else if (bpp == 24) {
                    if ((b->mode[idx].color & MASK_COLOR) == COLOR_YUVPL) {
                        planar422p_to_rgb24((const unsigned char *) b->m_buf[idx].frame[drv_frame].pdata, 
                                            (unsigned char *)ximage->data, 
                                            lpbmi->bmiHeader.biWidth,
                                            lpbmi->bmiHeader.biHeight);
                    }
                    else if ((b->mode[idx].color & MASK_COLOR) == COLOR_Y8) {
                        grayscale_rgb24((const unsigned char *) b->m_buf[idx].frame[drv_frame].pdata,  (unsigned char *)ximage->data, lpbmi->bmiHeader.biWidth, lpbmi->bmiHeader.biHeight);                                                
                    }
                }
                else if (bpp == 32) {
                    if ((b->mode[idx].color & MASK_COLOR) == COLOR_YUVPL) {
                        planar422p_to_rgb32((const unsigned char *) b->m_buf[idx].frame[drv_frame].pdata, 
                                            (unsigned char *)ximage->data, 
                                            lpbmi->bmiHeader.biWidth,
                                            lpbmi->bmiHeader.biHeight);
                    }
                    else if ((b->mode[idx].color & MASK_COLOR) == COLOR_Y8) {
                        grayscale_rgb32((const unsigned char *) b->m_buf[idx].frame[drv_frame].pdata,  (unsigned char *)ximage->data, lpbmi->bmiHeader.biWidth, lpbmi->bmiHeader.biHeight);                                                
                    }
                }
                else {
                    printf("display mode not supported yet. try snapshots.\n");
                }
            }
            else {
                // snapshots
                if ((b->mode[idx].color & MASK_COLOR) == COLOR_YUVPL) {
                    planar422p_to_rgb24(b->m_buf[idx].frame[drv_frame].pdata, 
                                        rgb_frame,
                                        lpbmi->bmiHeader.biWidth,
                                        lpbmi->bmiHeader.biHeight);
                    info.stride = lpbmi->bmiHeader.biWidth * 3;
                    info.width = lpbmi->bmiHeader.biWidth;
                    info.height = lpbmi->bmiHeader.biHeight;
                    // save the image
                    save_image_uncompressed( rgb_frame, b->fname, &info, 0);
                    printf("Took snapshot %s on channel %d\n", b->fname, idx+1);                    
                    prompt();
                    b->b_snapshot[idx] = 0;
                }
                else if ((b->mode[idx].color & MASK_COLOR) == COLOR_Y8) {
                    grayscale_rgb24(b->m_buf[idx].frame[drv_frame].pdata, 
                                    rgb_frame,
                                    lpbmi->bmiHeader.biWidth,
                                    lpbmi->bmiHeader.biHeight);
                    info.stride = lpbmi->bmiHeader.biWidth * 3;
                    info.width = lpbmi->bmiHeader.biWidth;
                    info.height = lpbmi->bmiHeader.biHeight;
                    // save the image
                    save_image_uncompressed( rgb_frame, b->fname, &info, 0);
                    printf("Took snapshot %s on channel %d\n", b->fname, idx+1);
                    prompt();
                    b->b_snapshot[idx] = 0;
                }
            }
            // Now we write the image to the pixmap.
            XShmPutImage(display, pix, gc, ximage, 0, 0, 0, 0, width, height,
                         True);
            // Copy the pixmap to the window
            XCopyArea(display, pix, windowId, gc, 0, 0, width, height, 0, 0);
            XNextEvent(display, &event);
            if (event.type == ClientMessage && 
                event.xclient.data.l[0] == wmDeleteMessage) {
                b->stop_rq[idx] = 1;
                printf("Window closed\n");
            }
            XSync(display, True);
        }
        // end work with frame

        // Done with frame, must now dequeue the buffer
        //printf("dq buf %p %d %d\n", b->handle, chn, drv_frame);
        S2255_DQBUF( b->handle, chn, drv_frame);
        active_frame++;
        if (active_frame >= NUMFRAMES) {
            active_frame = 0;
        }        
        b->frame_count[idx]++;        
    }

    // X windows, close the window
    if (display) {
        XFreePixmap(display, pix);
        closeWindow( display, windowId, gc);
        display = NULL;
    }

    prompt();

    if (b->smode[idx] == MODE_REC) {
        fclose( curfile);
    }
    
    // Stop the stream acquisition
    if (S2255_StopAcquire( b->handle, chn ) != 0) {
        printf("stop failed!");
    }

#ifndef COPY_METHOD
    // unmap mmap buffers
    for( i=0;i<NUMFRAMES;i++) {
        if (b->m_buf[idx].frame[i].pdata ) {
            //printf("munmap  pdata %p\n", b->m_buf[idx].frame[i].pdata);
            munmap( b->m_buf[idx].frame[i].pdata, size);
            b->m_buf[idx].frame[i].pdata = NULL;
        }            
    }
    b->m_buf[idx].size = 0;        
#else
    // free buffers
    b->m_buf[idx].size = 0;        
    for( i=0;i<NUMFRAMES;i++) {
        if (b->m_buf[idx].frame[i].pdata ) {
            free( b->m_buf[idx].frame[i].pdata );
            b->m_buf[idx].frame[i].pdata = NULL;
        }            
    }
#endif
    // report statistics.
    printf( "%u frames transferred. %u frames dropped\n", b->frame_count[idx], b->drop_count[idx] );
    prompt();
    free(rgb_frame);
    // free bitmap info header
    free(lpbmi);
    // Indicate stream is stopped.
    b->smode[idx] = MODE_IDLE;

    prompt();
    // Terminate thread.
    free( arg);
    printf( "Stream stopped.\n");

    /* if recording, change color mode back*/
    if (work_image) {
        free(work_image);
        work_image = NULL;
    }
    return 0;
}

// creates the X window
static Window createWindow(Display** display, GC *gc, int width, int height,int board, int chn)
{
    Window window;
    int majorv;
    int minorv;
    Bool pixmaps;
    XGCValues gcValues;

    *display = XOpenDisplay (NULL);
    if (!display)
    {
        fprintf (stderr, "createWindow: Can not open display\n");
        exit(1);
    }

    int iscreen = DefaultScreen(*display);
    unsigned long int white = XWhitePixel(*display, iscreen);
    unsigned long int black = XBlackPixel(*display, iscreen);

    if ((XShmQueryVersion (*display, &majorv, &minorv, &pixmaps) == 0)
        || (majorv < 1) || ((majorv == 1) && (minorv < 1))) 
    {
        fprintf (stderr, "createWindow: No xshm extension\n");
        exit(1);
    }

    window = XCreateSimpleWindow(*display, DefaultRootWindow(*display), 0, 0,
                                 width, height, 0, white, black);

    if (!window)
    {
        fprintf(stderr, "createWindow: Window could not be created.\n");
        exit(1);
    }

    gcValues.foreground = white;
    gcValues.background = black;
    gcValues.line_width = 1;

    *gc = XCreateGC (*display, window, 0, &gcValues);

    XMapRaised(*display, window);

    // add window name
    XTextProperty title;
    char *window_title;
    window_title = (char *) malloc( 100);
    if (window_title != NULL) {
        sprintf(window_title, "Sensoray 2255 [board %d(0 indexed)] channel: %d", board,chn);
        int rc = XStringListToTextProperty(&window_title,1,&title);
        if (rc == 0) {
            free(window_title);
            return window;
        }
    }
    XSetWMName( *display, window, &title);
    free(window_title);
    return window;
}

// closes the X window
void closeWindow(Display* display, Window win, GC gc)
{
    XFreeGC (display, gc);
    XDestroyWindow (display, win);
    XCloseDisplay (display);
}

