/*
 * 2255 USB Linux driver
 *
 * Copyright (C) 2007-2019 Sensoray Company Inc.
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License as
 *      published by the Free Software Foundation, version 2.
 *
 * @author: Dean A.
 * @date: 2007-2023
 */
/*
 * A note on Video for Linux (V4L or V4L2):
 * This driver was designed with custom IOCTLs similar to our 611 driver
 * Some users may prefer V4L or this driver
 * Both drivers are open source, but the Sensoray V4L driver was 
 * designed to be included into the Linux kernel.
 * This driver uses custom IOCTLs and will not be brought into the kernel
 * tree.
 */


#include "s2255core.h"
#include "s2255mod.h"
#include "s2255ioctl.h"
#include "size.h"
#include "f2255usb.h"
#include "2255vx.h"

#define KERNEL_VERSION_IOCTL_REMOVED       KERNEL_VERSION(2, 6, 36)
#define KERNEL_VERSION_USBCOMPLETE_CHANGED KERNEL_VERSION(2, 6, 19)
#define KERNEL_VERSION_MUTEX_ADDED         KERNEL_VERSION(2, 6, 16)
#define KERNEL_VERSION_VMRESERVED_GONE     KERNEL_VERSION(3, 7, 0)

#if LINUX_VERSION_CODE >= KERNEL_VERSION_MUTEX_ADDED
#define s2255_mutex_init     mutex_init
#define s2255_mutex_lock     mutex_lock
#define s2255_mutex_lock_interruptible mutex_lock_interruptible
#define s2255_mutex_unlock   mutex_unlock
#else
#define s2255_mutex_init     init_MUTEX
#define s2255_mutex_lock     down
#define s2255_mutex_lock_interruptible     down_interruptible
#define s2255_mutex_unlock   up
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION_VMRESERVED_GONE
#define VM_RESERVED (VM_DONTDUMP | VM_DONTEXPAND)
#endif

static int s2255_mmap(struct file *file, struct vm_area_struct *vma);
void s2255_sleep( int ms);

#define DRIVERNAME "s2255"

// vendor ids
#define USB_S2255_VENDOR_ID   0x1943
#define USB_S2255_PRODUCT_ID  0x2255
#define USB_S2257_PRODUCT_ID  0x2257


/* table of devices that work with this driver */
static struct usb_device_id s2255_table [] = {
	{ USB_DEVICE(USB_S2255_VENDOR_ID, USB_S2255_PRODUCT_ID) },
	{ USB_DEVICE(USB_S2255_VENDOR_ID, USB_S2257_PRODUCT_ID) },
	{ }                                     /* Terminating entry */
};
MODULE_DEVICE_TABLE (usb, s2255_table);

/* Get a minor range for your devices from the usb maintainer */
#define USB_S2255_MINOR_BASE    0

static int debug = 1;
int *s2255_debug=&debug;
module_param(debug,int,0);
MODULE_PARM_DESC(debug, "Debug level(0-4)");

#define dprintk(level,fmt, arg...)					\
	do {								\
		if (*s2255_debug >= (level)) {				\
			printk(KERN_DEBUG "s2255: " fmt, ##arg);	\
		}							\
	} while (0)


#define to_s2255_dev(d) container_of(d, struct s2255_dev, kref)


static struct usb_driver s2255_driver;

#if LINUX_VERSION_CODE >= KERNEL_VERSION_MUTEX_ADDED
static struct mutex s2255_devices_lock;
#else
static struct semaphore s2255_devices_lock;
#endif

struct s2255_dev { 
	PVOID      hbd; // board handle to core function for common driver functions across OSes
	int frames;
	int users; // file handles open
	int debug;
#if LINUX_VERSION_CODE >= KERNEL_VERSION_MUTEX_ADDED
	struct mutex lock;
#else
	struct semaphore lock;
#endif
	int                   resources;
	struct usb_device     *udev;			 // our USB device
	struct usb_interface  *interface;	 // usb interface

	__u8			        bulk_in_endpointAddr;	// address of bulk in endpoint
	__u8			        bulk_out_endpointAddr;	// address of bulk out endpoint

	struct kref		        kref;
	wait_queue_head_t	    frame_wq[MAX_CHANNELS];
	spinlock_t              frame_slock[MAX_CHANNELS];
	int                     frame_count[MAX_CHANNELS];
	int                     cur_input;
	int                     cur_fmt[2];
	int                     active_frame;
	BUFFER                  workbuf[MAX_CHANNELS];
	wait_queue_head_t	wait_setmode[MAX_CHANNELS];
	wait_queue_head_t	wait_fw;
	int                     vidstatus_ready[MAX_CHANNELS];
	int                     setmode_ready[MAX_CHANNELS];
	int                     fw_ready;
	wait_queue_head_t	wait_vidstatus[MAX_CHANNELS];
	char                    *usbbuf;
#if LINUX_VERSION_CODE >= KERNEL_VERSION_MUTEX_ADDED
	struct mutex            cmd_lock;
	struct mutex            ioctl_lock;
#else
	struct semaphore        cmd_lock;
	struct semaphore        ioctl_lock;
#endif
};
#define MAX_USBXFER_SIZE 64


struct s2255_fh {
	struct s2255_dev	*dev;
	unsigned int resources;	/* resource management for device open */
	int width;
	int height;
};

LONG s2255_vendor_request(PVOID pdx, 
			     UCHAR Request, 
			     UINT16 Index,
			     UINT16 Value,
			     PVOID TransferBuffer,
			     UINT32 TransferBufferLength,
			  int bOut);
static int s2255_get_fx2fw(struct s2255_dev *dev, int *fw);

int s2255_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);

long s2255_ioctl_unlocked(struct file *file,
			  unsigned int cmd, unsigned long arg)
{
        struct s2255_fh  *fh = (struct s2255_fh *) file->private_data;
	struct s2255_dev *pdx = (struct s2255_dev *) fh->dev;
	int rc;
	dprintk(4, "ioctl unlocked\n");
	if (s2255_mutex_lock_interruptible(&pdx->ioctl_lock)) {
	  printk("could not acquire lock s2226_open\n");
	  return -EAGAIN;
	}
	rc = s2255_ioctl(NULL, file, cmd, arg);
	s2255_mutex_unlock(&pdx->ioctl_lock);
	return rc;
}

// io control function
int s2255_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct s2255_fh  *fh = (struct s2255_fh *) file->private_data;
	struct s2255_dev *pdx = (struct s2255_dev *) fh->dev;
	
	void __user *argp = (void __user *)arg;
	int ret=0;
	dprintk(4, "ioctl\n");
	if (pdx == NULL) {
		printk("invalid device\n");
		return -EINVAL;
	}
	if (pdx->hbd == NULL) {
		printk("invalid device\n");
		return -EINVAL;
	}    

	switch( cmd) {
	case S2255_IOC_DQBUF:
        {
		dqbuf_param_t cmd_chnidx;
		if (argp == NULL)
			return -EINVAL;
		if (copy_from_user( &cmd_chnidx, argp, sizeof( cmd_chnidx)))
			return -EINVAL;
        if (cmd_chnidx.chn < 0 || cmd_chnidx.chn >= MAX_CHANNELS)
            return -EINVAL;
		dprintk(3,"dqbuf %d %d\n",cmd_chnidx.chn, cmd_chnidx.frm);
		ret = S2255CORE_dqbuf(pdx->hbd, cmd_chnidx.chn, cmd_chnidx.frm);
        }
        break;
	case S2255_IOC_SET_MODE:
        {
		set_mode_param_t cmd_mode;
		dprintk(4,"set mode\n");
		if (copy_from_user( &cmd_mode, argp, sizeof(set_mode_param_t)))
			return -EINVAL;
        if (cmd_mode.idx < 0 || cmd_mode.idx >= MAX_CHANNELS)
            return -EINVAL;
		dprintk(3,"set mode: idx %d\n", cmd_mode.idx);
		ret = S2255CORE_set_mode( pdx->hbd, cmd_mode.idx, &cmd_mode.mode);
		dprintk(3,"set mode: ret: %d\n",ret);
		break;
        }
	case S2255_IOC_GET_MODE:
        {
		set_mode_param_t cmd;
		dprintk(3, "s2255: get mode\n");
		if (copy_from_user( &cmd, argp, sizeof(set_mode_param_t)))
			return -EINVAL;
        if (cmd.idx < 0 || cmd.idx >= MAX_CHANNELS)
            return -EINVAL;
		dprintk(3,"get mode: idx %d\n", cmd.idx);
		ret = S2255CORE_get_mode( pdx->hbd, cmd.idx, &cmd.mode);
		if (ret != 0)
			return -EINVAL;
		if (copy_to_user( argp, &cmd, sizeof(cmd)))
			return -EINVAL;
		ret = 0;
		break;
        }
        // set buffer NOT for Linux
	case S2255_IOC_COPY_BUFFER:
        {
		getbuf_param_t  cmd;
		dprintk(3,"copy buffer\n");
		if (copy_from_user( &cmd, argp, sizeof(getbuf_param_t)))
			return -EINVAL;
        if (cmd.idx < 0 || cmd.idx >= MAX_CHANNELS)
            return -EINVAL;
		ret = S2255CORE_copy_buffer( pdx->hbd, cmd.idx, cmd.frm_num, cmd.pbuf, cmd.size);
		break;
        }
	case S2255_IOC_REQBUF:
        {
		getbuf_param_t  cmd;
		dprintk(3,"get buffer\n");
		if (argp == NULL) {
			printk("invalid arg\n");
			return -EINVAL;
		}
		if (copy_from_user( &cmd, argp, sizeof(getbuf_param_t)))
			return -EINVAL;
        if (cmd.idx < 0 || cmd.idx >= MAX_CHANNELS)
            return -EINVAL;
		cmd.pbuf = S2255CORE_get_buffer( pdx->hbd, cmd.idx,cmd.frm_num);
		dprintk(4,"got buffer 2: %p\n", cmd.pbuf);
		if (copy_to_user( argp, &cmd, sizeof(cmd)))
			return -EINVAL;
		dprintk(3,"CHn %d frm %d, first buffer %p\n", cmd.idx, cmd.frm_num, cmd.pbuf);
		ret = 0;
		break;
        }

	case S2255_IOC_STARTACQUIRE:
        {
		start_param_t cmd;
		dprintk(3,"S2255: start acquire\n");
		if (copy_from_user( &cmd, argp, sizeof( cmd)))
			return -EINVAL;
        if (cmd.idx < 0 || cmd.idx >= MAX_CHANNELS)
            return -EINVAL;
		dprintk(3,"start acquire channel %d\n", cmd.idx);
		//printk("start acquire %d\n",cmd.idx);
		ret = S2255CORE_start_acquire( pdx->hbd, cmd.idx);
		return 0;
		break;
        }
	case S2255_IOC_STOPACQUIRE:
        {
		stop_param_t cmd;
		if (copy_from_user( &cmd, argp, sizeof( cmd)))
			return -EINVAL;
        if (cmd.idx < 0 || cmd.idx >= MAX_CHANNELS)
            return -EINVAL;
		dprintk(3,"stop acquire channel %d\n", cmd.idx);
		ret = S2255CORE_stop_acquire( pdx->hbd, cmd.idx);
		break;
        }
	case S2255_IOC_GET_FRAME_COUNT:
        {
		UINT32 frms;
		frms = S2255CORE_get_frame_count( pdx->hbd);
		if (copy_to_user( argp, &frms, sizeof( frms)))
			return -EINVAL;
		return 0;
        }
	case S2255_IOC_WAIT_FRAME_NONBLOCK:
        {
		wait_param_t cmd;
		unsigned long flags;
		int count;
		if (copy_from_user( &cmd, argp, sizeof( cmd)))
			return -EINVAL;
		//printk("wait frame channel nonblock %d\n", cmd.idx);            
		if (cmd.idx >= MAX_CHANNELS || (cmd.idx < 0)) {
			return -EINVAL;
		}
		dprintk(4,"poll for frame, channel: %d count %d\n",cmd.idx, pdx->frame_count[cmd.idx]);
		spin_lock_irqsave( &pdx->frame_slock[cmd.idx], flags);
		count = pdx->frame_count[cmd.idx];
		spin_unlock_irqrestore( &pdx->frame_slock[cmd.idx], flags);
            
		if (count == 0) {
			dprintk(4,"poll channel[not ready]: %d count %d\n",cmd.idx, pdx->frame_count[cmd.idx]);
			return -EWOULDBLOCK;
		}
		dprintk(4,"poll channel result: %d count %d\n",cmd.idx, pdx->frame_count[cmd.idx]);
		ret = 0;
		S2255CORE_getlastframe(pdx->hbd, 
                               cmd.idx,
                               &cmd.last_frame,
                               &cmd.last_size,
                               &cmd.last_ts);
		spin_lock_irqsave( &pdx->frame_slock[cmd.idx], flags);
		if (pdx->frame_count[cmd.idx] ) {
			pdx->frame_count[cmd.idx]--;
		}
		spin_unlock_irqrestore( &pdx->frame_slock[cmd.idx], flags);
		if (copy_to_user( argp, &cmd, sizeof(cmd)))
			return -EINVAL;
		return 0;
        }
	case S2255_IOC_WAIT_FRAME_BLOCK:
        {
		wait_param_t cmd;
		int wait_ms;
		if (copy_from_user( &cmd, argp, sizeof(cmd))) {
			return -EINVAL;
		}
		//printk("wait frame channel blocking: %d\n", cmd.idx);
		wait_ms = cmd.timeout;

		if (cmd.idx >= MAX_CHANNELS || cmd.idx < 0) {
			return -EINVAL;
		}
		dprintk(4,"block channel[before]: %d count %d\n",cmd.idx, pdx->frame_count[cmd.idx]);
		ret = wait_event_interruptible_timeout( pdx->frame_wq[cmd.idx], (pdx->frame_count[cmd.idx] > 0), msecs_to_jiffies(wait_ms));
		// retrieve the current frame in the core
		// if ret == 0, no timeout so frames are ready
		if (ret != 0) {
			unsigned long flags;
			dprintk(4,"block channel[success]:chn %d count %d\n",cmd.idx, pdx->frame_count[cmd.idx]);
			spin_lock_irqsave( &pdx->frame_slock[cmd.idx], flags);

			if (pdx->frame_count[cmd.idx] ) {
				pdx->frame_count[cmd.idx]--;
			}
			spin_unlock_irqrestore( &pdx->frame_slock[cmd.idx], flags);
			S2255CORE_getlastframe(pdx->hbd, cmd.idx, &cmd.last_frame,
                                   &cmd.last_size, &cmd.last_ts);
			//printk("last size %d\n", cmd.last_size);
			if (copy_to_user( argp, &cmd, sizeof(cmd)))
				return -EINVAL;
			return 0;
		}
		dprintk(8, "timeout\n");
		return -ETIME;
        }
	case S2255_IOC_GET_STATUS:
        {
		get_vidstatus_t cmd;
		UINT32 status;
		dprintk(3,"S2255: get vid status\n");
		if (copy_from_user( &cmd, argp, sizeof( cmd)))
			return -EINVAL;
        if (cmd.idx < 0 || cmd.idx >= MAX_CHANNELS)
            return -EINVAL;
		printk("get status\n");
		cmd.status = 0;
		ret = S2255CORE_get_status(pdx->hbd, cmd.idx, &status);
		printk("got status %x\n", status);
		cmd.status = status;
		if (ret != 0)
			return -EINVAL;
		if (copy_to_user( argp, &cmd, sizeof(cmd)))
			return -EINVAL;
		return 0;
        }
	case S2255_IOC_QUERYBUF:
        {
            querybuf_t cmd;
            dprintk(3,"S2255: query buf\n");
            if (copy_from_user( &cmd, argp, sizeof( cmd)))
                return -EINVAL;
            cmd.state = S2255CORE_query_buf(pdx->hbd, cmd.chn, cmd.frm);
            if (copy_to_user( argp, &cmd, sizeof(cmd)))
                return -EINVAL;
            return 0;
        }
	case S2255_IOC_QUERYBUF_EXT:
        {
            querybuf_ext_t cmd;
            dprintk(3,"S2255: query buf ext\n");
            if (copy_from_user( &cmd, argp, sizeof( cmd)))
                return -EINVAL;
            cmd.state = S2255CORE_query_buf_ext(pdx->hbd, cmd.chn, cmd.frm, &cmd.ts);
            if (copy_to_user( argp, &cmd, sizeof(cmd)))
                return -EINVAL;
            return 0;
        }
	case S2255_IOC_QUERYBUF_ALL:
        {
            querybuf_all_t cmd;
            int k;
            dprintk(3,"S2255: query buf all\n");
            if (copy_from_user( &cmd, argp, sizeof( cmd)))
                return -EINVAL;
            for (k = 0; k < SYS_FRAMES; k++) {
                cmd.state[k] = S2255CORE_query_buf_ext(pdx->hbd, cmd.chn, k, &cmd.ts[k]);
            }
            if (copy_to_user( argp, &cmd, sizeof(querybuf_all_t)))
                return -EINVAL;
            return 0;
        }
	case S2255_IOC_GET_FIRMWARE:
        {
		get_firmware_t cmd;
		dprintk(3,"S2255: get vid status\n");
		if (copy_from_user( &cmd, argp, sizeof( cmd)))
			return -EINVAL;
		ret = S2255CORE_get_dspfw(pdx->hbd, &cmd.dsp_fw);
		if (ret != 0)
			return -EINVAL;
		ret = s2255_get_fx2fw(pdx, &cmd.usb_fw);
		if (ret != 0)
			return -EINVAL;
		if (copy_to_user( argp, &cmd, sizeof(cmd)))
			return -EINVAL;
		return 0;
        }
	case SX2255_IOC_READ_I2C_BURST:
		{
		get_sn_param_t cmd;
		dprintk(3,"S2255: get serial number\n");
		if (copy_from_user( &cmd, argp, sizeof( cmd)))
			return -EINVAL;
		if (cmd.datalen > MAX_USBXFER_SIZE)
			return -EINVAL;		
		if (pdx->usbbuf == NULL)
			return -EFAULT;
		memset(pdx->usbbuf, 0, MAX_USBXFER_SIZE);
		ret = S2255CORE_read_i2c_burst( pdx->hbd, cmd.addr, cmd.subaddr, pdx->usbbuf, cmd.datalen );				
		if (ret <= 0)
			return -EINVAL;
		memcpy(cmd.data, pdx->usbbuf, cmd.datalen);
		if (copy_to_user( argp, &cmd, sizeof(cmd)))
			return -EINVAL;		
		return 0;
		}

	default:
		ret = -EINVAL;
		break;
	}
	return ret;
}


static void s2255_delete(struct kref *kref)
{       
	struct s2255_dev *dev = to_s2255_dev(kref);
	if (dev) {
		if (dev->hbd ) {
			dprintk(4,"shutdown %p\n", dev->hbd);
			S2255CORE_board_shutdown( dev->hbd);
			dev->hbd = NULL;
		}
		else { 
			printk("dev null\n");
		}
	}
	else { 
		printk("dev null\n");
		return;
	}
	if (dev->usbbuf)
		kfree(dev->usbbuf);
	usb_put_dev(dev->udev);
	kfree (dev);
}

struct s2255_mapping {
	void *pvirt; // virtual start
	unsigned long size; // size of it
	struct s2255_dev *pdx;

};

static void s2255_vm_open (struct vm_area_struct *vma)
{
	dprintk(4, "VMopen\n");
}

static void s2255_vm_close (struct vm_area_struct *vma)
{
	struct s2255_mapping *pmap = (struct s2255_mapping *) vma->vm_private_data;
	if (pmap == NULL) {
		return;
	}
	dprintk(4, "VMclose\n");
	// free the mapping
	kfree( pmap);
	vma->vm_private_data = NULL;
	return;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 26)
static struct page *s2255_vm_nopage(struct vm_area_struct *vma, unsigned long address, int *type)
{
	struct page *page = NOPAGE_SIGBUS;
	struct s2255_mapping *pmap = (struct s2255_mapping *) vma->vm_private_data;
	unsigned long offset;
	if (!pmap) {
		printk("no map!\n");
		return NOPAGE_OOM;
	}
	if (address > vma->vm_end) {
		printk("no remap!\n");
		return NOPAGE_SIGBUS;
	}
	// calculate the offset
	offset = (unsigned long) ((address - vma->vm_start) + pmap->pvirt);
	page = vmalloc_to_page((void *)  offset);
	if (!page) {
		printk("no page!\n");
		return NOPAGE_OOM;
	}
	get_page(page);
	if (type) {
		*type = VM_FAULT_MINOR;
	}
	dprintk(4,"got offset %lx %lx\n", vma->vm_start, vma->vm_pgoff);
	return page;
}
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 1, 0)
static int s2255_vm_fault(struct vm_fault *vmf)
#elif LINUX_VERSION_CODE < KERNEL_VERSION(4, 11, 0)
static int s2255_vm_fault(struct vm_area_struct *vma,
			  struct vm_fault *vmf)
#else
vm_fault_t s2255_vm_fault(struct vm_fault *vmf)
#endif
{
	struct page *page;
	unsigned long offset;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 11, 0)	
	struct s2255_mapping *pmap = (struct s2255_mapping *) vma->vm_private_data;
#else
	struct vm_area_struct *vma;
	struct s2255_mapping *pmap = (struct s2255_mapping *) vmf->vma->vm_private_data;
	vma = vmf->vma;
#endif
	

	if (!pmap) {
		printk("no map!\n");
		return VM_FAULT_OOM;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0)
	// calculate the offset
	offset = (unsigned long) (((char *) vmf->virtual_address - (char *)vma->vm_start) + pmap->pvirt);
#else
	offset = (unsigned long) (((char *) vmf->address - (char *)vma->vm_start) + pmap->pvirt);
#endif
	page = vmalloc_to_page((void *)  offset);
	if (!page) {
		printk("no page!\n");
		return VM_FAULT_OOM;
	}
	get_page(page);
	vmf->page = page;
	dprintk(4,"got offset %lx %lx\n", vma->vm_start, vma->vm_pgoff);
	return 0;
}
#endif

// virtual memory operations
static struct vm_operations_struct s2255_vm_ops = {
	.open = s2255_vm_open,
	.close = s2255_vm_close,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 26)
	.nopage = s2255_vm_nopage,
#else
	.fault = s2255_vm_fault,
#endif
};

// map kernel buffers to user space
static int s2255_mmap(struct file *file, struct vm_area_struct *vma)
{
	//struct s2255_fh  *fh = (struct s2255_fh *) file->private_data;
	//struct s2255_dev *pdx = (struct s2255_dev *) fh->dev;
	size_t size;
	void *addr;
	struct s2255_mapping *pmap;
	addr = (void *) (vma->vm_pgoff * PAGE_SIZE);
	//printk("addr %p, %ld %ld\n", addr, vma->vm_pgoff, vma->vm_pgoff * PAGE_SIZE );
	size = vma->vm_end - vma->vm_start;
	//printk("MMAP size %d",size);
	pmap = (struct s2255_mapping *) kmalloc( sizeof( struct s2255_mapping) , GFP_KERNEL);
	if (pmap == NULL) {
		return -ENOMEM;
	}
	memset( pmap, 0, sizeof(struct s2255_mapping));
	vma->vm_ops = &s2255_vm_ops;
	pmap->pvirt = addr;
	pmap->size = size;
#if LINUX_VERSION_CODE < KERNEL_VERSION(6,3,0)
	vma->vm_flags |= VM_RESERVED | VM_DONTEXPAND;
	vma->vm_flags |= VM_IO;
#else
	vm_flags_set(vma, VM_RESERVED | VM_IO | VM_DONTEXPAND);
#endif
	vma->vm_private_data = pmap;
	s2255_vm_open(vma);
	//printk("VMA PGOFF remap %ld\n", vma->vm_pgoff);
	return 0;
}


// when device is opened
static int s2255_open(struct inode *inode, struct file *file)
{
	struct s2255_dev *dev;
	struct s2255_fh  *fh = NULL;
	struct usb_interface *interface;
	int subminor;
	int retval = 0;
	dprintk(2,"s2255_open\n");

	subminor = iminor(inode);
	interface = usb_find_interface(&s2255_driver, subminor);

	if (!interface) {
		printk(KERN_ERR "%s - error, can't find device for minor %d", __FUNCTION__, subminor);
		retval = -ENODEV;
		goto exit_err;
	}
	dev = usb_get_intfdata(interface);
	if (!dev) {
		retval = -ENODEV;
		goto exit_err;
	}

	if (s2255_mutex_lock_interruptible(&s2255_devices_lock)) {
		printk("could not acquire lock s2255_open\n");
		return -ERESTARTSYS;
	}

    
	if (dev->users > 0)
	{
		printk("s2255: device already open.\n");
		s2255_mutex_unlock(&s2255_devices_lock);
		return -ENODEV;
	}
    
	// allocate file context data
	fh = kmalloc(sizeof(struct s2255_fh),GFP_KERNEL);
	if (NULL == fh) {
		s2255_mutex_unlock(&s2255_devices_lock);
		return -ENOMEM;
	}
	memset(fh,0,sizeof(struct s2255_fh));
	printk("dev %p\n", dev);
	file->private_data = fh;
	fh->dev            = dev;



	dev->users++;
	// open the core board (only meant for one OPEN device instance)
	if (dev->hbd) {
		if (S2255CORE_board_open( dev->hbd) != 0) {
			printk("failed to open board!\n");
			retval = -ENODEV;
			s2255_mutex_unlock(&s2255_devices_lock);
			goto exit_err;
		}
	}
	// increment our usage count for the device 
	kref_get(&dev->kref);
	s2255_mutex_unlock(&s2255_devices_lock);
	return 0;
exit_err:
	if (fh) {
		kfree( fh);
	}
	return retval;
}

// when device is released
static int s2255_release(struct inode *inode, struct file *file)
{
	struct s2255_fh         *fh = file->private_data;
	struct s2255_dev        *dev;
	dprintk(2, "s2255_release\n");
	if (s2255_mutex_lock_interruptible(&s2255_devices_lock)) {
		return -ERESTARTSYS;
	}

	//      int minor = iminor(inode);
	dprintk(4, "s2255_release 1\n");
	if (fh == NULL) {
		s2255_mutex_unlock( &s2255_devices_lock);
		return -ENODEV;
	}
	dprintk(4, "s2255_release 2\n");
	dev = fh->dev;
	dprintk(4, "s2255_release 3\n");
	if (dev == NULL) { 
		s2255_mutex_unlock( &s2255_devices_lock);
		return -ENODEV;
	}
	dprintk(4, "s2255_release 4\n");
	dprintk(4, "s2255_release 5 %p\n", dev->hbd);
	kfree (fh);
	dev->users--;
	if (dev->users == 0) {
		if (dev->hbd) {
			// close the core board


			dprintk(4, "s2255_release 7 %p\n", dev->hbd);        
			if (S2255CORE_board_close( dev->hbd) != 0) {
				printk("failed to close board!\n");
			}
		}
	}
	s2255_mutex_unlock( &s2255_devices_lock);
	// decrement the count on our device
	kref_put(&dev->kref, s2255_delete);
	return 0;
}



static struct file_operations s2255_fops = {
	.owner =    THIS_MODULE,
	.open =     s2255_open,
#if LINUX_VERSION_CODE < KERNEL_VERSION_IOCTL_REMOVED
	.ioctl =    s2255_ioctl,
#else
	.unlocked_ioctl =    s2255_ioctl_unlocked,
#endif
	.mmap =     s2255_mmap,
	.release =  s2255_release,
};

/* 
 * usb class driver info in order to get a minor number from the usb core,
 * and to have the device registered with the driver core
 */
static struct usb_class_driver s2255_class = {
	.name =         "s2255_%d",
	.fops =         &s2255_fops,
	.minor_base =   USB_S2255_MINOR_BASE,
};


static int s2255_load_firmware(void *obj);
// probe function.  when USB device is detected by system
static int s2255_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	struct s2255_dev *dev = NULL;
	struct usb_host_interface *iface_desc;
	struct usb_endpoint_descriptor *endpoint;
	int i;
	int retval = -ENOMEM;
	printk("s2255: probe\n");
	/* allocate memory for our device state and initialize it to zero */
	dev = (struct s2255_dev *) kmalloc(sizeof(*dev), GFP_KERNEL);
	if (dev == NULL) {
		printk(KERN_ERR "s2255: out of memory");
		goto error;
	}
	memset(dev,0,sizeof(*dev));
	kref_init(&dev->kref);
	dev->usbbuf = kmalloc(MAX_USBXFER_SIZE, GFP_KERNEL | GFP_DMA);
	if (dev->usbbuf == NULL) {
		printk(KERN_ERR "s2255: out of memory");
		goto error;
	}
	
	dev->udev = usb_get_dev(interface_to_usbdev(interface));
	dev->interface = interface;
	// set up the endpoint information 
	iface_desc = interface->cur_altsetting;
	printk("num endpoints %d\n" ,iface_desc->desc.bNumEndpoints);
	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		endpoint = &iface_desc->endpoint[i].desc;
		if (!dev->bulk_in_endpointAddr &&
		    ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
		     == USB_DIR_IN) &&
		    ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
		     == USB_ENDPOINT_XFER_BULK)) {
			// we found the bulk in endpoint */
			dev->bulk_in_endpointAddr = endpoint->bEndpointAddress;
		}
		if (!dev->bulk_out_endpointAddr &&
		    ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
		     == USB_DIR_OUT) &&
		    ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
		     == USB_ENDPOINT_XFER_BULK)) {
			/* we found the bulk out endpoint */
			dev->bulk_out_endpointAddr = endpoint->bEndpointAddress;
			printk("bulk out endpoint addr: %d\n", dev->bulk_out_endpointAddr);
                }
	}
	if (!(dev->bulk_in_endpointAddr && dev->bulk_out_endpointAddr)) {
		printk(KERN_ERR "Could not find both bulk-in and bulk-out endpoints");
		goto error;
	}
	// open the 2255 core function
	dev->hbd = S2255CORE_board_init( 2255, (PVOID) dev);
	if (dev->hbd == NULL) {
		printk("FAILED to open 2255 board!\n");
		goto error;
	}
	// save our data pointer in this interface device 
	usb_set_intfdata(interface, dev);
	dprintk(4, "after intfdata %p\n", dev->hbd);

	// initialize the frame wait queues
	for( i=0;i< MAX_CHANNELS; i++) {
		init_waitqueue_head(&dev->frame_wq[i]);
		init_waitqueue_head(&dev->wait_setmode[i]);
		init_waitqueue_head(&dev->wait_vidstatus[i]);
		spin_lock_init(&dev->frame_slock[i]);
		dev->frame_count[i] = 0;
	}
	init_waitqueue_head(&dev->wait_fw);
	// initialize locks 
	s2255_mutex_init(&dev->lock);
	s2255_mutex_init(&dev->cmd_lock);
	s2255_mutex_init(&dev->ioctl_lock);
	/* we can register the device now, as it is ready */
	dprintk(4,"before usb register video rigester intfdata %p\n", dev->hbd);
	retval = usb_register_dev(interface, &s2255_class);
	dprintk(4,"after usb register video rigester intfdata %p\n", dev->hbd);
	if (retval) {
		/* something prevented us from registering this driver */
		printk(KERN_ERR "Not able to get a minor for this device.");
		usb_set_intfdata(interface, NULL);
		goto error;
	}
	/* let the user know what node this device is now attached to */
	printk(KERN_INFO "USB 2255 device now attached to USBs2255-%d", interface->minor);

	usb_reset_device( dev->udev);

	s2255_start_readpipe( dev, INDEX_READ_PIPE);


	// DOWNLOAD the main firmware to the DSP.
	// NOT using Linux 2.6 "firmware loader" class because
	// it has not been stable between kernel versions.
	// Also, firmware loader forces user to have a kernel compiled
	// with "firmware class" in it.  Some distributions do not do this.
	// So we download the firmware from header file.  This will always work,
	// and is not a major drawback(except increasing binary) as new firmware will
	// always require a new driver download anyways.
	retval = s2255_load_firmware((PVOID) dev);

	if (retval != 0) {
		usb_set_intfdata(interface, NULL);
		// unregister the driver.
		usb_deregister_dev(interface, &s2255_class);
		goto error;
	}

	printk("S2255 successfully loaded\n");
	dprintk(4,"before probe done %p\n",dev->hbd);
	return 0;
error:
	kref_put(&dev->kref, s2255_delete);
	return retval;
}



// disconnect routine.  when board is removed physically or with rmmod
static void s2255_disconnect(struct usb_interface *interface)
{
	struct s2255_dev *dev = NULL;
	int minor = interface->minor;
	// lock to prevent s2255_open() from racing s2255_disconnect() 
	s2255_mutex_lock( &s2255_devices_lock);
	// close 2255 board core
	dev = usb_get_intfdata(interface);
	usb_set_intfdata(interface, NULL);
	// give back our minor 
	usb_deregister_dev(interface, &s2255_class);
	kref_put(&dev->kref, s2255_delete);
	s2255_mutex_unlock( &s2255_devices_lock);
	printk(KERN_INFO "USB s2255eton #%d now disconnected", minor);
}



static struct usb_driver s2255_driver = {
	.name =         "s2255",
	.probe =        s2255_probe,
	.disconnect =   s2255_disconnect,
	.id_table =     s2255_table,
};


static int __init usb_s2255_init(void)
{
	int result;
	// register this driver with the USB subsystem 
	result = usb_register(&s2255_driver);

	if (result) {
		printk(KERN_ERR "usb_register failed. Error number %d", result);
	}

	s2255_mutex_init( &s2255_devices_lock);
	dprintk(2, "s2255_init: done\n");
	return result;
}

static void __exit usb_s2255_exit(void)
{
	/* deregister this driver with the USB subsystem */
	usb_deregister(&s2255_driver);
}


module_init (usb_s2255_init);
module_exit (usb_s2255_exit);

MODULE_DESCRIPTION("Sensoray 2255 Linux driver: Version 1.1.39");
MODULE_LICENSE("GPL");
MODULE_VERSION("1.1.39");

int s2255_lock_buffer( char *pToLock, ULONG num_bytes, char **pLocked, PVOID *pMDL)
{
	// Linux, do nothing.  don't need to lock the buffer, because we are
	// using kernel buffers not user buffers
	return 0;
}

int s2255_unlock_buffer( PVOID pMDL)
{
	return 0;
}

void *s2255_alloc_mem( int size ) 
{
	void *res;
	res = vmalloc( size );
	return res;
}

void s2255_free_mem( void *obj )
{
	if (obj) {
		vfree( obj );
	}
	return;
}

void *s2255_alloc_dma_mem( int size ) 
{
	void *res;

	res = kmalloc( size, GFP_KERNEL );
	dprintk(4,"alloc mem dma %p\n", res);
	return res;
}


void s2255_free_dma_mem( void *obj )
{
	dprintk(4,"free mem dma %p\n",obj);
	if (obj) {
		kfree( obj );
	}
	return;
}



void s2255_memset( void *p, UINT32 length, UINT8 fill)
{
	memset( p, fill, length);
	return;
}



int s2255_memcpy(void *pDest, void *pSrc, UINT32 Length)
{
	(void) memcpy( pDest, pSrc, Length);
	return 0;
}


int  s2255_copy_kernel_to_user( char *pdest, char *psrc, int size)
{
	return copy_to_user( pdest, psrc, size);
}


int s2255_I2C_ReadRegister(PVOID pdx, UCHAR dev_addr, UINT16 reg_addr, UCHAR *reg_value )
{
	return 0;
}
int s2255_I2C_WriteRegister(PVOID pdx, UCHAR dev_addr, UINT16 reg_addr, UCHAR reg_value )
{
	return 0;
}

int s2255_I2C_WriteReg2(PVOID pdx, UCHAR dev_addr, UCHAR datalen, UCHAR *data )
{
	return 0;
}

int s2255_I2C_WriteReg3(PVOID pdx, UCHAR dev_addr, UCHAR datalen, UCHAR *data )
{
	return 0;
}

int s2255_I2C_ReadReg2(PVOID pdx, UCHAR dev_addr, UINT8 datalen, UCHAR *data )
{
	return 0;
}

int s2255_I2C_ReadRegisterBurst(PVOID pdx, UCHAR dev_addr, UINT16 reg_addr, UCHAR *data, UCHAR datalen ) 
{
     int rc;	
     rc = s2255_vendor_request(pdx, VX_I2CREAD, reg_addr, dev_addr >> 1, data, datalen, DIR_IN);
     return rc;
}

static int s2255_get_fx2fw(struct s2255_dev *dev, int *fw)
{
	int ret;
	unsigned char transBuffer[64];
	ret = s2255_vendor_request(dev, VX_FW, 0, 0, transBuffer, 2, DIR_IN);
	if (ret < 0) {
		dprintk(2, "get fw error: %x\n", ret);
        return ret;
    }
	*fw = transBuffer[0] + (transBuffer[1] << 8);
	dprintk(2, "Get FW %x %x\n", transBuffer[0], transBuffer[1]);
	return 0;
}

void s2255_sleep( int ms)
{
	wait_queue_head_t sleep_q;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,69)
	DEFINE_WAIT(wait);
#endif
	if (ms == 0) {
		schedule();
		return;
	}
	init_waitqueue_head( &sleep_q);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,69)
	interruptible_sleep_on_timeout( &sleep_q (ms *HZ) / 1000);
#else
	prepare_to_wait(&sleep_q,&wait, TASK_INTERRUPTIBLE);
	schedule_timeout((ms*HZ) / 1000);
	finish_wait( &sleep_q, &wait);
#endif
}





// also in s2255core.c
static ULONG G_chnmap[MAX_CHANNELS]= {3,2,1,0};

int s2255_wakeup( void *obj, int chn)
{
	unsigned long flags;
	int idx;
	struct s2255_dev       *pdx = (struct s2255_dev *) obj;
	// must map back to actual channel
	idx = G_chnmap[chn];
	// increment the frame count
	dprintk(2,"wakeup: count %d channel: %d\n", pdx->frame_count[idx],idx);
	spin_lock_irqsave( &pdx->frame_slock[idx], flags);
	pdx->frame_count[idx]++;
	spin_unlock_irqrestore( &pdx->frame_slock[idx], flags);
	wake_up( &pdx->frame_wq[idx]);
	return 0;
}

int s2255_reset_frames( void *obj, int chn)
{
	unsigned long flags;
	int idx;
	struct s2255_dev       *pdx = (struct s2255_dev *) obj;
	// must map back to actual channel
	idx = G_chnmap[chn];
	// reset frame counter to 0
	spin_lock_irqsave( &pdx->frame_slock[idx], flags);
	pdx->frame_count[idx] = 0;
	spin_unlock_irqrestore( &pdx->frame_slock[idx], flags);
	return 0;
}

/* begin USB */
// forward
void ReadPipeCompletion(struct urb *purb);

int config_device( PVOID pdxc )
{
	return 0;
}



VOID s2255_usb_reset(PVOID pdx)
{                           // ResetDevice
	int r;
	struct s2255_dev *pdxc = (struct s2255_dev *) pdx;
	r = usb_reset_device( pdxc->udev);
	if (!r) {
		printk(KERN_DEBUG "invalid\n");
	}
	return;
}


ULONG s2255_usb_getstatus(PVOID pdx)
{
	int r = 0;
	//    struct s2255_dev *pdxc = (struct s2255_dev *) pdx;
	//  r = usb_get_status( pdxc->udev, USB_RECIP_DEVICE, 0, dev->);
	return r;
}   




LONG s2255_abort_pipe(PVOID pdx,UINT32 PipeNum)
{
	return 0;
}

// reset USB pipe.  also resets data on host
LONG s2255_reset_pipe(PVOID pdx,UINT32 PipeNum)
{
	return 0;
}

LONG s2255_vendor_request(PVOID pdx, 
			     UCHAR Request, 
			     UINT16 Index,
			     UINT16 Value,
			     PVOID TransferBuffer,
			     UINT32 TransferBufferLength,
			     int bOut)
{
	int r;

	struct s2255_dev *dev = (struct s2255_dev *) pdx;
	if (dev == NULL) {
		printk("failed vr\n");
		return 0;
	}

	if (!bOut) {		
		r = usb_control_msg(dev->udev, usb_rcvctrlpipe(dev->udev, 0),
				    Request,
				    USB_TYPE_VENDOR | USB_RECIP_DEVICE| USB_DIR_IN, Value, Index, TransferBuffer,
				    TransferBufferLength, 5000);
	}
	else {
		r = usb_control_msg(dev->udev, usb_sndctrlpipe(dev->udev, 0),
				    Request, USB_TYPE_VENDOR | USB_RECIP_DEVICE,
				    Value, Index, TransferBuffer,
				    TransferBufferLength, 2000);
	}
	return r;	
}



LONG s2255_write_config(PVOID pdx, UINT8 *pBuffer, SINT32 size)
{
	int pipe;
	struct s2255_dev    *dev;
	int done;
	LONG retval = -1;
	//dbgprint2("writeconfiguration NO CALLBACK:%p %d\n",pdx, size);
	dev = (struct s2255_dev *) pdx;
	if (dev->udev) {
		pipe = usb_sndbulkpipe(dev->udev,2);
		retval = usb_bulk_msg( dev->udev, pipe, pBuffer, size, &done, 500);
	}
	return retval;
}


// start acquire. 
int s2255_start_readpipe(PVOID pdx, UINT32 PipeID)
{
	int pipe;
	int retval;
	int i;
	struct s2255_dev    *pdxc = (struct s2255_dev *) pdx;
	PVPAR v = (PVPAR ) pdxc->hbd;
	S2255_USB_PIPE_INFO* pipe_info = (S2255_USB_PIPE_INFO *) &v->UsbPipes[PipeID];
	pipe = usb_rcvbulkpipe(pdxc->udev,pdxc->bulk_in_endpointAddr);

	for(i=0;i<MAX_PIPE_BUFFERS;i++) {
		pipe_info = &v->UsbPipes[PipeID][i];
		if (PipeID != INDEX_READ_PIPE ) {
			printk("invalid pipe!\n");
			return -1;   
		}
		pipe_info->dwState = 1;
		pipe_info->dwBufIdx = (UINT32) i;
		pipe_info->priority_set = 0;
		pipe_info->stream_urb = usb_alloc_urb(0,GFP_KERNEL);
		//s2255_reset_event(pipe_info->hStopEvent);
		if (!pipe_info->stream_urb) {
			printk("ReadStream : Unable to alloc URB");
			return -ENOMEM;
		}
		// transfer buffer allocated in board_init

		usb_fill_bulk_urb(pipe_info->stream_urb, pdxc->udev,
				  pipe, 
				  pipe_info->pTransferBuffer,
				  pipe_info->cur_transfer_size, (usb_complete_t) ReadPipeCompletion, pipe_info);

		pipe_info->dwUrbSize = sizeof(pipe_info->stream_urb);
		retval = usb_submit_urb( pipe_info->stream_urb, GFP_KERNEL);
		if (retval) {
			printk("START PIPE FAILED!\n");
			return retval;
		}
	}
	return 0;
}

void ReadPipeCompletion(struct urb *purb)
{
	S2255_USB_PIPE_INFO* pipe_info = (S2255_USB_PIPE_INFO*) purb->context;
	struct s2255_dev *pdxc = (PVOID) pipe_info->pdx;
	int status;
	int pipe;
	//printk("read pipe completion %p\n", purb);
	status = purb->status;
	//printk("status %d , %d\n", status, purb->status);
    /* if shutting down, return immediately */
	if (status == -ESHUTDOWN ) {
		//s2255_set_event(pipe_info->hStopEvent);
		printk("READPIPECOMPLETION: ERR\n");
		printk("urb cancelled readpipe\n");        
		return;
	}

	if (pipe_info->dwState == 0) {
		wake_up(&pipe_info->hStopEvent);
		printk("Exit USB pipe");
		return; // STATUS_MORE_PROCESSING_REQUIRED; 
	}

	if ((pipe_info->pCallBack != NULL) && (status == 0)) {
		pipe_info->pCallBack(pdxc->hbd, pipe_info);
	}

	pipe_info->dwErrorCount = 0;

	pipe = usb_rcvbulkpipe(pdxc->udev,pdxc->bulk_in_endpointAddr);
	// reuse urb
	usb_fill_bulk_urb(purb, pdxc->udev,
			  pipe, //usb_sndbulkpipe(dev->udev, dev->bulk_out_endpointAddr),
			  pipe_info->pTransferBuffer, pipe_info->cur_transfer_size, (usb_complete_t) ReadPipeCompletion, pipe_info);  

	if (pipe_info->dwState != 0 ) {
		if (usb_submit_urb( pipe_info->stream_urb, GFP_ATOMIC)) {
			printk("error submitting urb\n");
			usb_free_urb(pipe_info->stream_urb);
		}
	}
	return;
}


void s2255_stop_readpipe(PVOID pdx)
{
	struct s2255_dev *dev = (struct s2255_dev *) pdx;
	PVPAR v;
	UINT32 idx = INDEX_READ_PIPE;
	int j;
	if (dev == NULL) {
		printk(KERN_ERR "invalid device\n");
		return;
	}
	v = (PVPAR) dev->hbd;

	if (v == NULL) {
		printk(KERN_ERR "invalid board\n");
		return;
	}

	if (v->dbgaddr != &v->UsbPipes[0][0]) {
		printk(KERN_ERR "probably a fatal compile error %p!\n",
		       v->dbgaddr);
	}
	for( j=0;j<MAX_PIPE_BUFFERS;j++) {
		S2255_USB_PIPE_INFO *pipe_info = &v->UsbPipes[idx][j];
		if (pipe_info) {
			if (pipe_info->dwState == 0) {
				continue;
			}
			pipe_info->dwState = 0;
			pipe_info->dwPrevState = 1;

		}
		else {
			printk("invalid pipe\n");
			return;
		}
	}
	s2255_abort_pipe( pdx, idx);
	s2255_reset_pipe( pdx, idx);

	for( j=0;j< MAX_PIPE_BUFFERS;j++) {
		S2255_USB_PIPE_INFO* pipe_info = &v->UsbPipes[idx][j];
		if (pipe_info->dwPrevState == 1) {
            pipe_info->dwPrevState = 0;
		}
	}

	for( j=0;j<MAX_PIPE_BUFFERS;j++) {
		S2255_USB_PIPE_INFO* pipe_info = &v->UsbPipes[idx][j];
		if (pipe_info->stream_urb) {
			// cancel URB
			usb_kill_urb( pipe_info->stream_urb);
			usb_free_urb(pipe_info->stream_urb);
			pipe_info->stream_urb = NULL;
		}
	}
	return;
}


#define S2255_SETMODE_TIMEOUT 1000
int s2255_wait_setmode( void *obj, int chn)
{
	int idx;
	struct s2255_dev       *pdx = (struct s2255_dev *) obj;
	// must map back to actual channel
	idx = G_chnmap[chn];
	// reset frame counter to 0
	pdx->setmode_ready[idx] = 0;

	wait_event_timeout(pdx->wait_setmode[idx],
			   (pdx->setmode_ready[idx] != 0),
			   msecs_to_jiffies(S2255_SETMODE_TIMEOUT));
	if (pdx->setmode_ready[idx] != 1)
		printk(KERN_DEBUG "s2255: no set mode response\n");

	return 0;
}

int s2255_set_setmode( void *obj, int chn)
{
	int idx = chn; /* already reversed in core */
	struct s2255_dev *pdx = (struct s2255_dev *) obj;
	pdx->setmode_ready[idx] = 1;
	wake_up(&pdx->wait_setmode[idx]);
	return 0;
}


int s2255_set_fwready( void *obj)
{
	struct s2255_dev *pdx = (struct s2255_dev *) obj;
	pdx->fw_ready = 1;
	wake_up(&pdx->wait_fw);
	return 0;
}

int s2255_wait_vid_status( void *obj, int chn)
{
	int idx;
	struct s2255_dev       *pdx = (struct s2255_dev *) obj;
	// must map back to actual channel
	idx = G_chnmap[chn];
	// reset frame counter to 0
	pdx->vidstatus_ready[idx] = 0;
	wait_event_timeout(pdx->wait_vidstatus[idx],
			   (pdx->vidstatus_ready[idx] != 0),
			   msecs_to_jiffies(S2255_SETMODE_TIMEOUT));
	if (pdx->vidstatus_ready[idx] != 1) {
		printk(KERN_DEBUG "s2255: no vid status response\n");
		return -EFAULT;
	}
	return 0;
}
int s2255_set_vidstatus( void *obj, int chn)
{
	int idx = chn;
	struct s2255_dev *pdx = (struct s2255_dev *) obj;
	pdx->vidstatus_ready[idx] = 1;
	wake_up(&pdx->wait_vidstatus[idx]);
	return 0;
}


int s2255_load_firmware(void *obj)
{
	struct s2255_dev *dev = (struct s2255_dev *) obj;
	int retries = 0;
	unsigned char *pmem;
	unsigned long i, size;
#define CHUNK_SIZE 512
	// all USB transfers must be done with continuous kernel memory.
	// can't allocate more than 128k in current linux kernel, so
	// upload the firmware in chunks
	dev->fw_ready = 0;
#define S2255_FW_TIMEOUT 5000
	size = sizeof(G_f2255usb) / sizeof(unsigned char);
	pmem = kmalloc( CHUNK_SIZE, GFP_KERNEL);
	if (pmem == NULL) {
		printk("OUT OF MEMORY!\n");
		return -ENOMEM;
	}

#define MAX_RETRIES 6
	while (retries < MAX_RETRIES) {
		for(i = 0; i < size; i += CHUNK_SIZE) {
			int usb_ret;
			memcpy( pmem, G_f2255usb + i, (i+CHUNK_SIZE > size) ? (size % CHUNK_SIZE ) : CHUNK_SIZE);
			usb_ret = s2255_write_config(dev, pmem,  CHUNK_SIZE);
			if (usb_ret != 0) {
				break;
			}
		}
		if (i >= size) {
			// device may take some time to load firmware and start		
			wait_event_timeout(dev->wait_fw,
					   (dev->fw_ready != 0),
					   msecs_to_jiffies(S2255_FW_TIMEOUT));
		}
		
		if (dev->fw_ready != 1) {
			printk(KERN_ERR "s2255: firmware didn't load. retrying %d\n", retries);
			s2255_vendor_request(dev, 0x40, 0x0000, 0x0001, NULL, 0, 1);
			s2255_sleep(10);
			s2255_vendor_request(dev, 0x50, 0x0000, 0x0000, NULL, 0, 1);
			s2255_sleep(600);
			s2255_vendor_request(dev, 0x10, 0x0000, 0x0000, NULL, 0, 1);
		} else 
			break;
		retries++;
	}

	kfree(pmem);
	
	if (dev->fw_ready != 1) {
		printk("S2255: USB download failed.  replug in device and reload driver(rmmod s2255 first)\n");
		return -ENOMEM;
	}
	return 0;
}

int s2255_lock_command(void *pdx)
{
	struct s2255_dev *dev = (struct s2255_dev *) pdx;
	s2255_mutex_lock(&dev->cmd_lock);
	return 0;
}

int s2255_unlock_command(void *pdx)
{
	struct s2255_dev *dev = (struct s2255_dev *) pdx;
	s2255_mutex_unlock(&dev->cmd_lock);
	return 0;
}
