/*
 *  V4L2 video capture example
 *
 *  This program can be used and distributed without restrictions.
 *  Tab set to 4 (driver code set to 8)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <getopt.h>             /* getopt_long() */
#include <fcntl.h>              /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <malloc.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <asm/types.h>          /* for videodev2.h */
#include <linux/videodev2.h>
#include <linux/media.h>

#define MAX_MEDIA_DEVICES 100
struct media_device_info mdi;
static void test_video_status(void);
#define CLEAR(x) memset (&(x), 0, sizeof (x))

typedef enum {
	IO_METHOD_READ,
	IO_METHOD_MMAP,
	IO_METHOD_USERPTR,
} io_method;

struct buffer {
    void *                  start;
    size_t                  length;
};

static char dev_name[100];
static io_method	io		= IO_METHOD_MMAP;
static int              fd              = -1;
struct buffer *         buffers         = NULL;
static unsigned int     n_buffers       = 0;
static int G_interp = 0;
static int G_serialnum = 0;
static int G_fwver = 0;

volatile int g_stopping;
void sigHandler(int type)
{
	g_stopping = 1;
	return;
}

static void
errno_exit                      (const char *           s)
{
    fprintf (stderr, "%s error %d, %s\n",
             s, errno, strerror (errno));

    exit (EXIT_FAILURE);
}

static int
xioctl                          (int                    fd,
                                 int                    request,
                                 void *                 arg)
{
    int r;

    do r = ioctl (fd, request, arg);
    while (-1 == r && EINTR == errno);

    return r;
}

#define V4L2_CID_S2255_COLORFILTER (V4L2_CID_USER_S2255_BASE + 0)
#define V4L2_CID_S2255_DEVICEID (V4L2_CID_USER_S2255_BASE + 1) // serial number
#define V4L2_CID_S2255_DSPFWVER (V4L2_CID_USER_S2255_BASE + 2)
#define V4L2_CID_S2255_USBFWVER (V4L2_CID_USER_S2255_BASE + 3)

static int get_ctrl(int fd, int idx, int *val)
{
	int rc;
	struct v4l2_control ctrl;
	memset(&ctrl, 0, sizeof(struct v4l2_control));
	ctrl.id = idx;
	rc = xioctl(fd, VIDIOC_G_CTRL, &ctrl);
	if (rc == 0)
		*val = ctrl.value;
	return rc;
}

static int get_media_info(int fd, struct media_device_info *mdi)
{
	int rc;
	rc = xioctl(fd, MEDIA_IOC_DEVICE_INFO, mdi);
	return rc;
}

static void
process_image                   (const void *           p)
{
    fputc ('.', stdout);
    fflush (stdout);
}

static int
read_frame			(void)
{
    struct v4l2_buffer buf;
	unsigned int i;

	switch (io) {
	case IO_METHOD_READ:
        if (-1 == read (fd, buffers[0].start, buffers[0].length)) {
            switch (errno) {
            case EAGAIN:
                return 0;

			case EIO:
				/* Could ignore EIO, see spec. */

				/* fall through */

			default:
				errno_exit ("read");
			}
		}

        process_image (buffers[0].start);

		break;

	case IO_METHOD_MMAP:
		CLEAR (buf);

        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_MMAP;

        if (-1 == xioctl (fd, VIDIOC_DQBUF, &buf)) {
            switch (errno) {
            case EAGAIN:
                return 0;

			case EIO:
				/* Could ignore EIO, see spec. */

				/* fall through */

			default:
				errno_exit ("VIDIOC_DQBUF");
			}
		}

        assert (buf.index < n_buffers);

        process_image (buffers[buf.index].start);

		if (-1 == xioctl (fd, VIDIOC_QBUF, &buf))
			errno_exit ("VIDIOC_QBUF");

		break;

	case IO_METHOD_USERPTR:
		CLEAR (buf);

        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_USERPTR;

		if (-1 == xioctl (fd, VIDIOC_DQBUF, &buf)) {
			switch (errno) {
			case EAGAIN:
				return 0;

			case EIO:
				/* Could ignore EIO, see spec. */

				/* fall through */

			default:
				errno_exit ("VIDIOC_DQBUF");
			}
		}

		for (i = 0; i < n_buffers; ++i)
			if (buf.m.userptr == (unsigned long) buffers[i].start
			    && buf.length == buffers[i].length)
				break;

		assert (i < n_buffers);

        process_image ((void *) buf.m.userptr);

		if (-1 == xioctl (fd, VIDIOC_QBUF, &buf))
			errno_exit ("VIDIOC_QBUF");

		break;
	}

	return 1;
}


static void
mainloop                        (void)
{
	unsigned int count;

    count = 100;

    while (count-- > 0 && !g_stopping) {
        for (;;) {
            fd_set fds;
            struct timeval tv;
            int r;

            FD_ZERO (&fds);
            FD_SET (fd, &fds);

            /* Timeout. */
            tv.tv_sec = 2;
            tv.tv_usec = 0;

            r = select (fd + 1, &fds, NULL, NULL, &tv);

            if (-1 == r) {
                if (EINTR == errno)
                    continue;

                errno_exit ("select");
            }

            if (0 == r) {
                fprintf (stderr, "select timeout\n");
                exit (EXIT_FAILURE);
            }

			if (read_frame ())
                break;
	
			/* EAGAIN - continue select loop. */
        }
    }
}

static void
stop_capturing                  (void)
{
    enum v4l2_buf_type type;

	switch (io) {
	case IO_METHOD_READ:
		/* Nothing to do. */
		break;

	case IO_METHOD_MMAP:
	case IO_METHOD_USERPTR:
		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

		if (-1 == xioctl (fd, VIDIOC_STREAMOFF, &type))
			errno_exit ("VIDIOC_STREAMOFF");

		break;
	}
}

static void
start_capturing                 (void)
{
    unsigned int i;
    enum v4l2_buf_type type;

	switch (io) {
	case IO_METHOD_READ:
		/* Nothing to do. */
		break;

	case IO_METHOD_MMAP:
		for (i = 0; i < n_buffers; ++i) {
            struct v4l2_buffer buf;

            CLEAR (buf);

            buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory      = V4L2_MEMORY_MMAP;
            buf.index       = i;

            if (-1 == xioctl (fd, VIDIOC_QBUF, &buf))
                errno_exit ("VIDIOC_QBUF");
		}
		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

		if (-1 == xioctl (fd, VIDIOC_STREAMON, &type))
			errno_exit ("VIDIOC_STREAMON");

		break;

	case IO_METHOD_USERPTR:
		for (i = 0; i < n_buffers; ++i) {
            struct v4l2_buffer buf;

            CLEAR (buf);

            buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory      = V4L2_MEMORY_USERPTR;
			buf.index       = i;
			buf.m.userptr	= (unsigned long) buffers[i].start;
			buf.length      = buffers[i].length;

			if (-1 == xioctl (fd, VIDIOC_QBUF, &buf))
                errno_exit ("VIDIOC_QBUF");
		}

		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

		if (-1 == xioctl (fd, VIDIOC_STREAMON, &type))
			errno_exit ("VIDIOC_STREAMON");

		break;
	}
}

static void
uninit_device                   (void)
{
    unsigned int i;

	switch (io) {
	case IO_METHOD_READ:
		free (buffers[0].start);
		break;

	case IO_METHOD_MMAP:
		for (i = 0; i < n_buffers; ++i)
			if (-1 == munmap (buffers[i].start, buffers[i].length))
				errno_exit ("munmap");
		break;

	case IO_METHOD_USERPTR:
		for (i = 0; i < n_buffers; ++i)
			free (buffers[i].start);
		break;
	}

	free (buffers);
}

static void
init_read			(unsigned int		buffer_size)
{
    buffers = calloc (1, sizeof (*buffers));

    if (!buffers) {
        fprintf (stderr, "Out of memory\n");
        exit (EXIT_FAILURE);
    }

	buffers[0].length = buffer_size;
	buffers[0].start = malloc (buffer_size);

	if (!buffers[0].start) {
        fprintf (stderr, "Out of memory\n");
        exit (EXIT_FAILURE);
	}
}

static void
init_mmap			(void)
{
	struct v4l2_requestbuffers req;

    CLEAR (req);

    req.count               = 4;
    req.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory              = V4L2_MEMORY_MMAP;

	if (-1 == xioctl (fd, VIDIOC_REQBUFS, &req)) {
        if (EINVAL == errno) {
            fprintf (stderr, "%s does not support "
                     "memory mapping\n", dev_name);
            exit (EXIT_FAILURE);
        } else {
            errno_exit ("VIDIOC_REQBUFS");
        }
    }

    if (req.count < 2) {
        fprintf (stderr, "Insufficient buffer memory on %s\n",
                 dev_name);
        exit (EXIT_FAILURE);
    }

    buffers = calloc (req.count, sizeof (*buffers));

    if (!buffers) {
        fprintf (stderr, "Out of memory\n");
        exit (EXIT_FAILURE);
    }

    for (n_buffers = 0; n_buffers < req.count; ++n_buffers) {
        struct v4l2_buffer buf;

        CLEAR (buf);

        buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory      = V4L2_MEMORY_MMAP;
        buf.index       = n_buffers;

        if (-1 == xioctl (fd, VIDIOC_QUERYBUF, &buf))
            errno_exit ("VIDIOC_QUERYBUF");

        buffers[n_buffers].length = buf.length;
        buffers[n_buffers].start =
            mmap (NULL /* start anywhere */,
                  buf.length,
                  PROT_READ | PROT_WRITE /* required */,
                  MAP_SHARED /* recommended */,
                  fd, buf.m.offset);

        if (MAP_FAILED == buffers[n_buffers].start)
            errno_exit ("mmap");
    }
}

static void
init_userp			(unsigned int		buffer_size)
{
	struct v4l2_requestbuffers req;
    unsigned int page_size;

    page_size = getpagesize ();
    buffer_size = (buffer_size + page_size - 1) & ~(page_size - 1);

    CLEAR (req);

    req.count               = 4;
    req.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory              = V4L2_MEMORY_USERPTR;

    if (-1 == xioctl (fd, VIDIOC_REQBUFS, &req)) {
        if (EINVAL == errno) {
            fprintf (stderr, "%s does not support "
                     "user pointer i/o\n", dev_name);
            exit (EXIT_FAILURE);
        } else {
            errno_exit ("VIDIOC_REQBUFS");
        }
    }

    buffers = calloc (4, sizeof (*buffers));

    if (!buffers) {
        fprintf (stderr, "Out of memory\n");
        exit (EXIT_FAILURE);
    }

    for (n_buffers = 0; n_buffers < 4; ++n_buffers) {
        buffers[n_buffers].length = buffer_size;
        buffers[n_buffers].start = memalign (/* boundary */ page_size,
                                             buffer_size);

        if (!buffers[n_buffers].start) {
            fprintf (stderr, "Out of memory\n");
            exit (EXIT_FAILURE);
		}
    }
}
int find_media_for_video(int idx);
static void
init_device                     (int idx)
{
    struct v4l2_capability cap;
    struct v4l2_cropcap cropcap;
    struct v4l2_crop crop;
    struct v4l2_format fmt;
	struct v4l2_streamparm sp;
	unsigned int min;



    if (-1 == xioctl (fd, VIDIOC_QUERYCAP, &cap)) {
        if (EINVAL == errno) {
            fprintf (stderr, "%s is no V4L2 device\n",
                     dev_name);
            exit (EXIT_FAILURE);
        } else {
            errno_exit ("VIDIOC_QUERYCAP");
        }
    }

    if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
        fprintf (stderr, "%s is no video capture device\n",
                 dev_name);
        exit (EXIT_FAILURE);
    }

	switch (io) {
	case IO_METHOD_READ:
		if (!(cap.capabilities & V4L2_CAP_READWRITE)) {
			fprintf (stderr, "%s does not support read i/o\n",
                     dev_name);
			exit (EXIT_FAILURE);
		}

		break;

	case IO_METHOD_MMAP:
	case IO_METHOD_USERPTR:
		if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
			fprintf (stderr, "%s does not support streaming i/o\n",
                     dev_name);
			exit (EXIT_FAILURE);
		}

		break;
	}


    /* Select video input, video standard and tune here. */


	CLEAR (cropcap);

    cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (0 == xioctl (fd, VIDIOC_CROPCAP, &cropcap)) {
        crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        crop.c = cropcap.defrect; /* reset to default */

        if (-1 == xioctl (fd, VIDIOC_S_CROP, &crop)) {
            switch (errno) {
            case EINVAL:
                /* Cropping not supported. */
                break;
            default:
                /* Errors ignored. */
                break;
            }
        }
    } else {	
        /* Errors ignored. */
    }

	/* if interp */
	sp.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	sp.parm.capture.capturemode = G_interp ? V4L2_MODE_HIGHQUALITY : 0;
	if (-1 == xioctl (fd, VIDIOC_S_PARM, &sp)) {
		if (EINVAL == errno) {
			fprintf (stderr, "high quality mode not supported\n");
			exit (EXIT_FAILURE);
		} else {
			errno_exit ("VIDIOC_S_PARM");
		}
	} else {
		printf("set interp mode\n");
	}

    CLEAR (fmt);

    fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fmt.fmt.pix.width       = 640; 
    fmt.fmt.pix.height      = 480;
    fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
    fmt.fmt.pix.field       = V4L2_FIELD_INTERLACED;

    if (-1 == xioctl (fd, VIDIOC_S_FMT, &fmt))
        errno_exit ("VIDIOC_S_FMT");



    /* Note VIDIOC_S_FMT may change width and height. */

	/* Buggy driver paranoia. */
	min = fmt.fmt.pix.width * 2;
	if (fmt.fmt.pix.bytesperline < min)
		fmt.fmt.pix.bytesperline = min;
	min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
	if (fmt.fmt.pix.sizeimage < min)
		fmt.fmt.pix.sizeimage = min;

	switch (io) {
	case IO_METHOD_READ:
		init_read (fmt.fmt.pix.sizeimage);
		break;

	case IO_METHOD_MMAP:
		init_mmap ();
		break;

	case IO_METHOD_USERPTR:
		init_userp (fmt.fmt.pix.sizeimage);
		break;
	}
	test_video_status();

    // serial number retrieval example
    if (G_serialnum) {
	    int rc;
	    int sn;

#if 0
        struct s2255_sn {
            unsigned int sn;
            unsigned char month;
            unsigned char day;
            unsigned char year;
        };
        struct s2255_sn sn;
#define S2255_IOC_GET_SN _IOR('v', BASE_VIDIOC_PRIVATE+0, struct s2255_sn)
        if (xioctl (fd, S2255_IOC_GET_SN, &sn) == 0) {
            printf("Serial Number: %d [%02d-%02d-%02d]\n", sn.sn, sn.month,
                   sn.day, sn.year);
        } else {
            printf("could not get serial number or older board without SN\n");
        }
#endif
	int media_fd;
	media_fd = find_media_for_video(idx);
	if (media_fd != -1) {
		rc = get_media_info(media_fd, &mdi);
		if (rc >= 0)
			printf("MEDIA_DEVICE: name: %s, serial_num: %s\n", mdi.driver, mdi.serial);
		else
			printf("MEDIA_DEVICE: error retrieving %d\n", rc);
		close(media_fd);
	} else {
		printf("No associated media device found\n");
	}

	rc = get_ctrl(fd, V4L2_CID_S2255_DEVICEID, &sn);
	if (rc == 0) {
		printf("Serial Number: %d\n", sn);
        } else {
		printf("could not get serial number or older board without SN\n");
	}
    }
    // firmware retrieval
    if (G_fwver) {
	    int rc;
	    int dsp, usb;
	    int fail = 0;
#if 0
        struct s2255_fw {
            unsigned int usb;
            unsigned int dsp;
        };
        struct s2255_fw fw;
#define S2255_IOC_GET_FW _IOR('v', BASE_VIDIOC_PRIVATE+1, struct s2255_fw)
        if (xioctl (fd, S2255_IOC_GET_FW, &fw) == 0) {
			printf("USB fw %d.%d, DSP fw %d.%02d.%02d\n",
                   (fw.usb >> 8) & 0xff, fw.usb & 0xff,
                   fw.dsp / 10000, (fw.dsp % 10000) / 100, (fw.dsp % 100));
        } else {
            printf("could not get firmware version\n");
        }
#endif
	rc = get_ctrl(fd, V4L2_CID_S2255_DSPFWVER, &dsp);
	if (rc != 0) {
		printf("could not get DSP firmware version\n");
		fail++;
	}
	rc = get_ctrl(fd, V4L2_CID_S2255_USBFWVER, &usb);
	if (rc != 0) {
		printf("could not get USB firmware version\n");
		fail++;
	}
	if (fail == 0)
		printf("USB fw %d.%d, DSP fw %d.%02d.%02d\n",
			(usb >> 8) & 0xff, usb & 0xff,
			dsp / 10000, (dsp % 10000) / 100, (dsp % 100));
    }
}

static void
close_device                    (void)
{
    if (-1 == close (fd))
        errno_exit ("close");

    fd = -1;
}

static void
open_device                     (void)
{
    struct stat st; 

    if (-1 == stat (dev_name, &st)) {
        fprintf (stderr, "Cannot identify '%s': %d, %s\n",
                 dev_name, errno, strerror (errno));
        exit (EXIT_FAILURE);
    }

    if (!S_ISCHR (st.st_mode)) {
        fprintf (stderr, "%s is no device\n", dev_name);
        exit (EXIT_FAILURE);
    }

    fd = open (dev_name, O_RDWR /* required */ | O_NONBLOCK, 0);

    if (-1 == fd) {
        fprintf (stderr, "Cannot open '%s': %d, %s\n",
                 dev_name, errno, strerror (errno));
        exit (EXIT_FAILURE);
    }
}

static void
usage                           (FILE *                 fp,
                                 int                    argc,
                                 char **                argv)
{
    fprintf (fp,
             "Usage: %s [options]\n\n"
             "Options:\n"
             "-d | --device name   Video device name. if unspecified, searches for first 2255 channel\n"
             "-e | --enumerate     Enumerates all the 2255s in the system.\n"
             "-h | --help          Print this message\n"
             "-m | --mmap          Use memory mapped buffers\n"
             "-r | --read          Use read() calls\n"
             "-u | --userp         Use application allocated buffers\n"
             "-i | --interp        Interpolate full size fields 4CIFS\n"
             "-s | --serial        Retrieve serial number(if available)\n"
             "",
             argv[0]);
}

static const char short_options [] = "d:ehmruisf";

static const struct option
long_options [] = {
    { "device",     required_argument,      NULL,           'd' },
    { "enumerate",  no_argument,            NULL,           'e' },
    { "help",       no_argument,            NULL,           'h' },
    { "mmap",       no_argument,            NULL,           'm' },
    { "read",       no_argument,            NULL,           'r' },
    { "userp",      no_argument,            NULL,           'u' },
    { "interp",     no_argument,            NULL,           'i' },
    { "serial",     no_argument,            NULL,           's' },
    { "firmware",     no_argument,            NULL,         'f' },
    { 0, 0, 0, 0 }
};

static const char *cardname = "s2255";

/* returns -1 if no 2255 boards found.
   Otherwise returns index of first board in system
*/
static int EnumVideo(int display)
{
	int i, fd, rc;
    int first_2255 = -1;
	char devname[16];
	struct v4l2_capability cap;

	for (i = 0; i < 256; i++) {
		sprintf(devname, "/dev/video%d", i);
		fd = open(devname, O_RDWR);
		if (fd == -1)
			continue;
		if (xioctl(fd, VIDIOC_QUERYCAP, &cap) == 0) {
			if (strcmp(cap.card, cardname) == 0) {
                if (first_2255 == -1)
                    first_2255 = i;
                if (display)
                    printf("%s is a 2255 channel\n", devname);
            }
		}
		close(fd);
	}
	return first_2255;
}


static int Check2255(char *devname)
{
	int fd, rc;
	struct v4l2_capability cap;
    rc = -1;
    fd = open(devname, O_RDWR);
    if (fd == -1)
        return -1;
    if (xioctl(fd, VIDIOC_QUERYCAP, &cap) == 0) {
        if (strcmp(cap.card, cardname) == 0)
            rc = 0;
	}
    close(fd);
    return rc;

}

int find_media_for_video(int idx)
{
	int i;
	int fd;
	char medianame[64];
	int rc;
	struct media_entity_desc mdesc;

	memset(&mdesc, 0, sizeof(struct media_entity_desc));
	for (i = 0; i < MAX_MEDIA_DEVICES; i++) {
		sprintf(medianame, "/dev/media%d", i);
		fd = open(medianame, O_RDWR | O_NONBLOCK, 0);
		if (fd == -1)
			break;
		do {
			mdesc.id |= MEDIA_ENT_ID_FLAG_NEXT;
			rc = xioctl(fd, MEDIA_IOC_ENUM_ENTITIES, &mdesc);
			if (rc < 0)
				break;
			if (mdesc.dev.minor == idx)
				return fd;
//			printf("name: %s, id %d. raw: %d.%d\n",
//				mdesc.name, mdesc.id,
//				mdesc.dev.major, mdesc.dev.minor);
		} while (rc >= 0);
		close(fd);
	}
	return fd;
}

int
main                            (int                    argc,
                                 char **                argv)
{
    int idx;
    int dev_specified = 0;
    int enumerate=0;
    strcpy(dev_name, "/dev/video0");
    signal(SIGINT, &sigHandler);
    for (;;) {
        int index;
        int c;

        c = getopt_long (argc, argv,
                         short_options, long_options,
                         &index);

        if (-1 == c)
            break;


        switch (c) {
        case 0: /* getopt_long() flag */
            break;
        case 'd':
            strcpy(dev_name, optarg);
            dev_specified = 1;
            break;

        case 'e':
            enumerate = 1;
            break;

        case 'h':
            usage (stdout, argc, argv);
            exit (EXIT_SUCCESS);

        case 'm':
            io = IO_METHOD_MMAP;
			break;

        case 'r':
            io = IO_METHOD_READ;
			break;
        case 'u':
            io = IO_METHOD_USERPTR;
			break;
	case 'i':
		printf("using interpolation for full size\n");
		G_interp = 1;
		break;
        case 's':
            G_serialnum = 1;
            break;
        case 'f':
            G_fwver = 1;
            break;
        default:
            usage (stderr, argc, argv);
            exit (EXIT_FAILURE);
        }
    }
    idx = EnumVideo(enumerate);
    if (idx == -1) {
        printf("no 2255s found in system\n");
        exit (EXIT_FAILURE);
    }
    if (enumerate)
        exit (0);
    if (!dev_specified) {
        sprintf(dev_name, "/dev/video%d", idx);
        printf("No device specified using first 2255 channel found: %s\n", dev_name);
    }

    if (Check2255(dev_name)) {
        printf("%s is not a 2255 device! Use -e to enumerate boards\n", dev_name);
        exit (EXIT_FAILURE);
    }

    open_device ();

    init_device (idx);
    if (G_serialnum || G_fwver)
	    goto skip_capture;

    start_capturing ();
    mainloop ();
    stop_capturing ();
skip_capture:
    uninit_device ();

    close_device ();

    exit (EXIT_SUCCESS);

    return 0;
}


/*
 * demonstrates testing the video status to see 
 * if input signal connected or not.
 */
static void test_video_status()
{
	struct v4l2_input inp;
    if (-1 == xioctl (fd, VIDIOC_ENUMINPUT, &inp)) {
		printf("could not get video input status\n");
		return;
	}
	if (inp.status & V4L2_IN_ST_NO_SIGNAL)
		printf("no signal connected\n");
	else
		printf("video signal connected\n");
	return;
}
